'''
Created on 20 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from warnings import warn

from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import box, Point
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class LatLonLib(object):
    '''
    classdocs
    '''
    ALGER = GeoDataFrame([{'geometry': Point((3.042048, 36.752887))}], crs='epsg:4326')
    CAP_TOWN = GeoDataFrame([{'geometry': Point((18.42322, -33.92584))}], crs='epsg:4326')
    DAKAR = GeoDataFrame([{'geometry': Point((-17.444061, 14.6937))}], crs='epsg:4326')
    EDINBURGH = GeoDataFrame([{'geometry': Point((-3.188267, 55.953252))}], crs='epsg:4326')
    LAGOS = GeoDataFrame([{'geometry': Point((3.379206, 6.524379))}], crs='epsg:4326')
    # outside the polar circles ; latitude must be between -66.56 and +66.56
    # LONGYEARBYEN = GeoDataFrame([{'geometry': Point((-15.6258907, 78.2253594))}], crs='epsg:4326')
    MONTPELLIER = GeoDataFrame([{'geometry': Point((3.87723, 43.61092))}], crs='epsg:4326')
    NANTES = GeoDataFrame([{'geometry': Point((-1.55, 47.2))}], crs='epsg:4326')
    NIAMEY = GeoDataFrame([{'geometry': Point((2.1098, 13.51366))}], crs='epsg:4326')
    QUITO = GeoDataFrame([{'geometry': Point((-78.52495, -0.22985))}], crs='epsg:4326')
    REYKJAVIK = GeoDataFrame([{'geometry': Point((-21.89541, 64.13548))}], crs='epsg:4326')

    @staticmethod
    def fromGeoDataFrameToLatLon(gdf=NANTES):
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, 'GeoDataFrame')

        if (gdf.crs is None):
            warn('Crs is not set!')
            gdf = LatLonLib.NANTES

        _gdf = gdf if ('epsg:4326' == gdf.crs) else gdf.to_crs('epsg:4326')
        _centroid = box(*_gdf.total_bounds).centroid
        _latitude, _longitude = _centroid.y, _centroid.x
        return _latitude, _longitude



'''
Created on 16 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''


class ArrayCoding(object):
    '''
    classdocs
    '''

    @staticmethod
    def encode(listOfValues, separator='#'):
        try:
            iter(listOfValues)
            return separator.join([str(v) for v in listOfValues])
        except TypeError:
            return str(listOfValues)

    @staticmethod
    def decode(string, outputType=float, separator='#'):
        if (0 == len(string)):
            return []
        return [outputType(v) for v in string.split(separator)]



'''
Created on 5 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import pi
from pandas import isna
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class HeightOfRoughness(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, itemsGdf, elevationFieldName='HAUTEUR', buffDist=None):
        '''
        Constructor
        '''
        if not isinstance(itemsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(itemsGdf, 'GeoDataFrame')
        self.itemsGdf = itemsGdf
        self.spatialIndex = itemsGdf.sindex

        self.elevationFieldName = elevationFieldName
        self.buffDist = buffDist

    def runWithArgs(self, row):
        rowGeom, zone, zoneArea = row.geometry, None, 0.0

        if (self.buffDist is None) and (GeomLib.isPolygonal(rowGeom)):
            zone, zoneArea = rowGeom, rowGeom.area

        elif isinstance(self.buffDist, (int, float)):
            zone, zoneArea = rowGeom.centroid.buffer(self.buffDist), pi * self.buffDist ** 2

        if (0.0 < zoneArea):
            _features = GeomLib.extractFeaturesWithin(zone, self.itemsGdf, self.spatialIndex)
            _vol = sum([f['geometry'].area * f[self.elevationFieldName] for f in _features if not isna(f[self.elevationFieldName])])
            return { 'hre': float(_vol / zoneArea) }

        return { 'hre': None }



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''

from shapely.geometry import Point

from numpy import ceil, float64, floor


class Epsilon(object):
    '''
    classdocs
    '''
    EPSILON = 1E-6

    @staticmethod
    def equals(x, y, epsilon=EPSILON):
        # print 'compare:', x, y, type(x), type(y)
        if ((isinstance(x, int) or isinstance(x, float) or isinstance(x, float64)) and
            (isinstance(y, int) or isinstance(y, float) or isinstance(y, float64))):
            if (abs(x - y) <= epsilon):
                return True
        elif (isinstance(x, Point) and isinstance(y, Point)):
            if ((abs(x.x - y.x) < epsilon) and (abs(x.y - y.y) < epsilon)):
                return True
        elif (isinstance(x, list) or isinstance(x, tuple)) and (isinstance(y, list) or isinstance(y, tuple)):
            if len(x) == len(y):
                return all([Epsilon.equals(x[i], y[i], epsilon) for i in range(len(x))])
        return False

    @staticmethod
    def isANonDecimalValue(value, epsilon=EPSILON):
        aa = abs(value)
        aaa = aa - int(aa)
        return Epsilon.isZero(aaa, epsilon) or Epsilon.isZero(aaa - 1, epsilon)

    @staticmethod
    def isAMultipleOf(a, b, epsilon=EPSILON):
        return Epsilon.isANonDecimalValue(a / b, epsilon)

    @staticmethod
    def isZero(x, epsilon=EPSILON):
        if (isinstance(x, int) or isinstance(x, float) or isinstance(x, float64)):
            return Epsilon.equals(x, 0, epsilon)
        elif isinstance(x, Point):
            return (Epsilon.equals(x.x, 0, epsilon) and Epsilon.equals(x.y, 0, epsilon))
        elif (isinstance(x, list) or isinstance(x, tuple)):
            return all([Epsilon.isZero(item, epsilon) for item in x])
        raise Exception('Illegal argument: x must be int, float, QgsPoint, list, or tuple (of int, float or QgsPoint)!')

    @staticmethod
    def round(x, decimals, flag):
        x, flag = x * 10 ** decimals, flag.upper()
        if 'UP' == flag:
            return ceil(x) * 10 ** (-decimals)
        elif 'DOWN' == flag:
            return floor(x) * 10 ** (-decimals)
        raise Exception('Illegal argument: flag must be "up" or "down"!')



'''
Created on 20 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import date, datetime, time, timedelta, timezone 

from numpy import ndarray
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class DatetimeLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def generate(dt, tz=timezone.utc):
        if ((isinstance(dt, (list, ndarray, tuple))) and (3 == len(dt)) and
            (isinstance(dt[0], datetime)) and (isinstance(dt[1], datetime)) and
            (isinstance(dt[2], timedelta))):
            _dtStart, _dtStop, _dtStep = dt
            # Timezone alignment
            _dtStart = datetime(*_dtStart.timetuple()[:6], tzinfo=tz)
            _dtStop = datetime(*_dtStop.timetuple()[:6], tzinfo=tz)
            _dt = _dtStart

            result = []
            while (_dt <= _dtStop):
                result.append(_dt)
                _dt = _dt + _dtStep
            return { '%s - %s' % (str(_dtStart), str(_dtStop)): result }

        elif isinstance(dt, datetime):
            return { str(dt): [dt.replace(tzinfo=tz)] }

        elif isinstance(dt, date):
            result = []
            _year, _month, _day = dt.timetuple()[:3]
            for _hour in range(4, 20):
                for _minute in range(0, 60, 30):
                    result.append(datetime(_year, _month, _day, _hour, _minute, tzinfo=tz))
            return { str(dt): result }

        elif isinstance(dt, time):
            result = []
            _year = date.today().year
            _tz = tz if (dt.tzinfo is None) else dt.tzinfo
            for _month in range(1, 13):
                for _day in range(1, 22, 10):
                    result.append(datetime(_year, _month, _day, dt.hour,
                                           dt.minute, dt.second,
                                           tzinfo=_tz))
            return { str(dt): result }

        elif isinstance(dt, (list, ndarray, tuple)):
            result = {}
            for _dt in dt:
                result.update(DatetimeLib.generate(_dt, tz))
            return result

        elif isinstance(dt, dict):
            return dt

        raise IllegalArgumentTypeException(dt, '(list of) date, time, datetime, or 3-uple of (datetime, datetime, timedelta)')

    @staticmethod
    def fromDatetimesDictToListOfSunPositions(datetimes, sunModel, tz=timezone.utc):
        if not isinstance(datetimes, dict):
            datetimes = DatetimeLib.generate(datetimes, tz)

        result = []

        for _lbl, _dts in datetimes.items():
            for _dt in _dts:
                radDir = sunModel.getRadiationDirection(_dt)
                solarAlti, solarAzim = sunModel.getSolarAnglesInRadians(_dt)
                result.append([_dt, radDir, solarAlti, solarAzim])       

        return result

    '''
    @staticmethod
    def fromListOfDatetimesToListOfSunPositions(datetimes, sunModel, tz=timezone.utc):
        result = []
        for dt in datetimes:
            radDir = sunModel.getRadiationDirection(dt)
            solarAlti, solarAzim = sunModel.getSolarAnglesInRadians(dt)
            result.append([dt, radDir, solarAlti, solarAzim])
        return result
    '''

    @staticmethod
    def range(dtStart, dtStop, dtDelta, tz=timezone.utc):
        if (isinstance(dtStart, (date, datetime)) and 
            isinstance(dtStop, (date, datetime)) and
            isinstance(dtDelta, timedelta)):

            if isinstance(dtStart, date):
                dtStart = datetime(*dtStart.timetuple()[:3], tzinfo=tz)
            elif dtStart.tzinfo is None:
                dtStart = dtStart.replace(tzinfo=tz)

            if isinstance(dtStop, date):
                dtStop = datetime(*dtStop.timetuple()[:3], 23, 59, tzinfo=tz)
            elif dtStop.tzinfo is None:
                dtStop = dtStop.replace(tzinfo=tz)

            result, _dt = [], dtStart
            while (_dt <= dtStop):
                result.append(_dt)
                _dt = _dt + dtDelta
            return result

        raise Exception('Expecting: 3-uple of (datetime, datetime, timedelta)')

    @staticmethod
    def shareTheSameTZ(dt1, dt2):
        return (
            isinstance(dt1, datetime) and 
            isinstance(dt2, datetime) and 
            (dt1.tzinfo == dt2.tzinfo)
            )



'''
Created on 28 avr. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import date, datetime, timedelta

from geopandas import GeoDataFrame
from numpy import dot
from shapely.geometry import Point
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.DatetimeLib import DatetimeLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.sun.SunLib import SunLib
from t4gpd.energy.PerrinDeBrichambaut import PerrinDeBrichambaut
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess

from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.sun.SunBeamLib import SunBeamLib


class DirectSolarIrradiation(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, sensorsGdf, masksGdf, sensorsNormalFieldname, maskElevationFieldname,
                 dtStart, dtStop, dtDelta=timedelta(hours=1), model='pysolar'):
        '''
        Constructor
        '''
        if not isinstance(sensorsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(sensorsGdf, 'GeoDataFrame')

        if sensorsNormalFieldname not in sensorsGdf:
            raise Exception(f'{sensorsNormalFieldname} is not a relevant field name!')
        self.sensorsNormalFieldname = sensorsNormalFieldname

        if not isinstance(masksGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(masksGdf, 'GeoDataFrame')
        self.masksGdf = masksGdf
        self.masksSIdx = masksGdf.sindex

        if maskElevationFieldname not in masksGdf:
            raise Exception(f'{maskElevationFieldname} is not a relevant field name!')
        self.maskElevationFieldname = maskElevationFieldname

        if not GeoDataFrameLib.shareTheSameCrs(sensorsGdf, masksGdf):
            raise Exception(f'The sensors & masks GeoDataFrame must share same CRS!')

        if not isinstance(dtStart, (date, datetime)):
            raise IllegalArgumentTypeException(dtStart, 'date or datetime')
        self.dt0 = dtStart

        if not isinstance(dtStop, (date, datetime)):
            raise IllegalArgumentTypeException(dtStop, 'date or datetime')
        self.dt1 = dtStop

        if not isinstance(dtDelta, timedelta):
            raise IllegalArgumentTypeException(dtDelta, 'timedelta')
        self.ddt = dtDelta

        self.model = model

        magnWh = dtDelta.total_seconds() / 3600
        datetimes = DatetimeLib.range(dtStart, dtStop, dtDelta)
        sunModel = SunLib(masksGdf, model)
        self.sunBeamsAndDni = DirectSolarIrradiation.__fromListOfDatetimesToListOfSunBeamsAndDNI(
            datetimes, magnWh, sunModel)

    @staticmethod
    def __fromListOfDatetimesToListOfSunBeamsAndDNI(datetimes, magnWh, sunModel, skyType=PerrinDeBrichambaut.STANDARD_SKY):
        result = []
        for dt in datetimes:
            radDir = sunModel.getRadiationDirection(dt)
            solarAlti, _ = sunModel.getSolarAnglesInRadians(dt)
            dni = PerrinDeBrichambaut.directNormalIrradiance(solarAlti, skyType) if (0.0 < solarAlti) else 0.0
            result.append([dt, radDir, dni * magnWh])
        return result

    def runWithArgs(self, row):
        geom = row.geometry
        if not isinstance(geom, Point):
            raise IllegalArgumentTypeException(geom, 'Point')
        normal = row[self.sensorsNormalFieldname]
        if isinstance(normal, str):
            normal = ArrayCoding.decode(normal)

        wh = 0
        for dt, radDir, dni in self.sunBeamsAndDni:
            if (0.0 < dni):
                dotProd = dot(normal, radDir)
                if (0 < dotProd):
                    hitBySunBeam, _ = SunBeamLib.isHitBySunBeam(geom, dt, self.masksGdf, self.maskElevationFieldname, self.model)
                    if hitBySunBeam:
                        wh += dni * dotProd
        return { 'direct_irradiation': wh }



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import isnan
from pandas.core.frame import DataFrame
from t4gpd.comfort.algo.TmrtOutLib import TmrtOutLib
from t4gpd.comfort.indices.AbstractThermalComfortIndice import AbstractThermalComfortIndice
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class TmrtOut(AbstractThermalComfortIndice):
    '''
    classdocs
    '''

    def __init__(self, sensorsGdf, N='N', SRUp='SR01Up_1_Avg', SRDn='SR01Dn_1_Avg',
                 IRUp='IR01UpCo_1_Avg', IRDn='IR01DnCo_1_Avg'):
        '''
        Constructor

        N:
        SRUp:
        SRDn:
        IRUp:
        IRDn:
        '''
        if not isinstance(sensorsGdf, DataFrame):
            raise IllegalArgumentTypeException(sensorsGdf, 'DataFrame')

        for fieldname in (N, SRUp, SRDn, IRUp, IRDn):
            if fieldname not in sensorsGdf:
                raise Exception('%s is not a relevant field name!' % fieldname)

        self.N = N
        self.SRUp = SRUp
        self.SRDn = SRDn
        self.IRUp = IRUp
        self.IRDn = IRDn

    def runWithArgs(self, row):
        N = row[self.N]
        SRUp = row[self.SRUp]
        SRDn = row[self.SRDn]
        IRUp = row[self.IRUp]
        IRDn = row[self.IRDn]

        if (N is None) or isnan(N):
            return { 'Tmrt_OUT': None }

        # Tmrt_OUT: outdoor mean radiant temperature
        Tmrt_OUT = TmrtOutLib.assess_tmrt_out(N, SRUp, SRDn, IRUp, IRDn)

        return { 'Tmrt_OUT': Tmrt_OUT }



'''
Created on 10 mars 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import mean, sqrt
from shapely.geometry import LinearRing, Point, Polygon
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class GeomLib3D(object):
    '''
    classdocs
    '''

    @staticmethod
    def centroid(geom):
        tmp = GeomLib.getListOfShapelyPoints(geom, withoutClosingLoops=True)
        if (0 == len(tmp)):
            return Point()
        tmp = [[p.x, p.y, p.z] for p in tmp]
        return Point(mean(tmp, axis=0))

    @staticmethod
    def crossProduct(u, v):
        return [
            (u[1] * v[2]) - (u[2] * v[1]),
            (u[2] * v[0]) - (u[0] * v[2]),
            (u[0] * v[1]) - (u[1] * v[0])
            ]

    @staticmethod
    def distFromTo(a, b):
        return sqrt(GeomLib3D.squaredDistance3D(a, b))

    @staticmethod
    def dotProduct(u, v):
        return ((u[0] * v[0]) + (u[1] * v[1]) + (u[2] * v[2]))

    @staticmethod
    def __getArea(ring):
        # ring is a 3D contour
        coords = ring.coords
        _first = coords[0]

        _sum = [0, 0, 0]
        for i in range(1, len(coords) - 1):
            _curr, _next = coords[i], coords[i + 1]
            _product = GeomLib3D.crossProduct(
                GeomLib3D.vector_to(_first, _curr),
                GeomLib3D.vector_to(_first, _next))
            for j in range(3):
                _sum[j] += _product[j]

        _normal = GeomLib3D.getFaceNormalVector(ring)

        return 0.5 * GeomLib3D.dotProduct(_sum, _normal)

    @staticmethod
    def getArea(geom):
        if not GeomLib.isPolygonal(geom):
            raise IllegalArgumentTypeException(geom, 'Polygon or MultiPolygon')

        if (not GeomLib.is3D(geom)):
            return geom.area

        polygons = [geom] if isinstance(geom, Polygon) else geom.geoms

        result = 0
        for polygon in polygons:
            result += GeomLib3D.__getArea(polygon.exterior)
            for hole in polygon.interiors:
                result += GeomLib3D.__getArea(hole)
        return result

    @staticmethod
    def getFaceNormalVector(geom):
        if (not isinstance(geom, (LinearRing, Polygon))):
            raise IllegalArgumentTypeException(geom, 'LinearRing or Polygon')

        geom = geom.exterior if isinstance(geom, Polygon) else geom

        if (not GeomLib.is3D(geom)):
            _ccw = 1 if GeomLib.isCCW(geom) else -1
            return [0, 0, _ccw]

        coords = geom.coords

        # Maximize radius starting from first node
        _maxDiam, _maxDiamIdx = -float('inf'), None
        for i in range(1, len(coords) - 1):
            _currDiam = GeomLib3D.squaredDistance3D(coords[0], coords[i])
            if (_maxDiam < _currDiam):
                _maxDiam, _maxDiamIdx = _currDiam, i

        _firstRadius = GeomLib3D.vector_to(coords[0], coords[_maxDiamIdx])

        # Maximize surface area (absolute) value using first radius
        _maxNorm, _maxCrossProduct, _secondRadius = 0.0, None, None
        for i in range(1, len(coords) - 1):
            if (i != _maxDiamIdx):
                _currRadius = GeomLib3D.vector_to(coords[0], coords[i])
                _currCrossProduct = GeomLib3D.crossProduct(_firstRadius, _currRadius)
                if (i < _maxDiamIdx):
                    _currCrossProduct = [-_currCrossProduct[0], -_currCrossProduct[1], -_currCrossProduct[2]]
                _currNorm = GeomLib3D.squaredNorm3D(_currCrossProduct)
                if (abs(_maxNorm) < abs(_currNorm)):
                    _maxNorm, _maxCrossProduct, _secondRadius = _currNorm, _currCrossProduct, _currRadius

        return GeomLib3D.unitVector(_maxCrossProduct)

    @staticmethod
    def norm3D(u):
        return sqrt(GeomLib3D.squaredNorm3D(u))

    @staticmethod
    def squaredDistance3D(p, q):
        dx, dy, dz = GeomLib3D.vector_to(p, q)
        return ((dx * dx) + (dy * dy) + (dz * dz))

    @staticmethod
    def squaredNorm3D(u):
        return ((u[0] * u[0]) + (u[1] * u[1]) + (u[2] * u[2]))

    @staticmethod
    def unitVector(u):
        _norm = GeomLib3D.norm3D(u)
        if (0 < _norm):
            _invNorm = 1.0 / _norm
            return [_invNorm * u[0], _invNorm * u[1], _invNorm * u[2]]
        raise IllegalArgumentTypeException(u, 'Must be non-null vector!')

    @staticmethod
    def vector_to(p, q):
        dx = q[0] - p[0]
        dy = q[1] - p[1]
        dz = q[2] - p[2]
        return [dx, dy, dz]



'''
Created on 1 juil. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import date, datetime, timedelta, timezone

from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class CalendarLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def getEncapsulatingWeek(givenDate=date.today()):
        '''
        Returns the first and last day of the week, i.e. the Monday and Sunday
        that include the given date
        '''
        if not isinstance(givenDate, (date, datetime)):
            raise IllegalArgumentTypeException(givenDate, 'date or datetime')

        _curr = datetime(*givenDate.timetuple()[:3], 12, tzinfo=timezone.utc) 
        _currWeekDay = _curr.weekday()
        _firstDayOfWeek = _curr - timedelta(days=_currWeekDay)
        _lastDayOfWeek = _curr + timedelta(days=6 - _currWeekDay)
        _firstDayOfWeek = date(*_firstDayOfWeek.timetuple()[:3])
        _lastDayOfWeek = date(*_lastDayOfWeek.timetuple()[:3])
        return _firstDayOfWeek, _lastDayOfWeek



'''
Created on 27 aug. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import array_equal, cos, cross, dot, eye, sin, vstack
from numpy.linalg import norm
from shapely.coords import CoordinateSequence
from shapely.geometry import GeometryCollection, MultiPoint, MultiLineString, \
    MultiPolygon, LinearRing, LineString, Point, Polygon
from shapely.geometry.base import GeometrySequence
from t4gpd.commons.GeomLib import GeomLib


class RotationLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def makeRotationMatrix(vectA, vectB):
        '''
        OCTAVE FORMALISM:
        U = [1; 0; 0];
        V = [0; 1; 0];
        I = eye(3);
        cosTheta = dot(U,V);
        sinTheta = norm(cross(U,V));
        a = cross(U,V) / sinTheta;
        R = cosTheta * I + (1 - cosTheta) * a * a' + sinTheta * cross(I, [a,a,a])'
        '''
        # epsilon = 1e-16  # TO PREVENT THE ZERO DIVISION
        epsilon = 0
        A = vectA / (norm(vectA) + epsilon)
        B = vectB / (norm(vectB) + epsilon)
        assert not array_equal(A, B), 'Normalized vectors must not be equal'

        # https://www.developpez.net/forums/d1235157/general-developpement/algorithme-mathematiques/mathematiques/matrice-rotation-entre-vecteurs/
        # https://fr.wikipedia.org/wiki/Rotation_vectorielle
        I = eye(3)
        cosTheta = dot(A, B)
        sinTheta = norm(cross(A, B))
        a = cross(A, B) / sinTheta

        R = cosTheta * I + (1 - cosTheta) * a.reshape(-1, 1) * a.reshape(1, -1) + \
            sinTheta * cross(I, vstack([a, a, a]).T, axisa=0, axisb=0, axisc=0).T

        return R

    '''
    @staticmethod
    def rotation(axis, p, angle):
        axisNorm = sqrt(axis[0] * axis[0] + axis[1] * axis[1] + axis[2] * axis[2])
        u = (axis[0] / axisNorm, axis[1] / axisNorm, axis[2] / axisNorm)
        c, s = cos(angle), sin(angle)
        m = (
            (u[0] * u[0] * (1 - c) + c, u[0] * u[1] * (1 - c) - u[2] * s, u[0] * u[2] * (1 - c) + u[1] * s),
            (u[0] * u[1] * (1 - c) + u[2] * s, u[1] * u[1] * (1 - c) + c, u[1] * u[2] * (1 - c) - u[0] * s),
            (u[0] * u[2] * (1 - c) - u[1] * s, u[1] * u[2] * (1 - c) + u[0] * s, u[2] * u[2] * (1 - c) + c)
            )
        if isinstance(p, Point):
            p = [p.x, p.y, p.z]
        return (
            m[0][0] * p[0] + m[0][1] * p[1] + m[0][2] * p[2],
            m[1][0] * p[0] + m[1][1] * p[1] + m[1][2] * p[2],
            m[2][0] * p[0] + m[2][1] * p[1] + m[2][2] * p[2]
            )

    @staticmethod
    def xAxisRotation(p, angle):
        return RotationLib.rotation((1, 0, 0), p, angle)

    @staticmethod
    def yAxisRotation(p, angle):
        return RotationLib.rotation((0, 1, 0), p, angle)

    @staticmethod
    def zAxisRotation(p, angle):
        return RotationLib.rotation((0, 0, 1), p, angle)
    '''

    @staticmethod
    def rotate2DXYZ(point, angleInRad):
        c, s = cos(angleInRad), sin(angleInRad)

        if isinstance(point, Point):
            if point.has_z:
                p = [point.x, point.y, point.z]
            else:
                p = [point.x, point.y]
        else:
            p = point

        x, y = p[0] * c - p[1] * s, p[0] * s + p[1] * c
        if (2 == len(p)):
            return [x, y]
        elif (3 == len(p)):
            return [x, y, p[2]]
        return None

    @staticmethod
    def rotate2D(geom, angleInRad):
        if isinstance(geom, CoordinateSequence):
            return [RotationLib.rotate2DXYZ(p, angleInRad) for p in geom]

        elif isinstance(geom, GeometrySequence):
            return [RotationLib.rotate2D(g, angleInRad) for g in geom]

        elif GeomLib.isAShapelyGeometry(geom):
            if isinstance(geom, Point):
                return Point(RotationLib.rotate2DXYZ(geom, angleInRad))
            elif isinstance(geom, LinearRing):
                return LinearRing(RotationLib.rotate2D(geom.coords, angleInRad))
            elif isinstance(geom, LineString):
                return LineString(RotationLib.rotate2D(geom.coords, angleInRad))
            elif isinstance(geom, Polygon):
                extRing = RotationLib.rotate2D(geom.exterior.coords, angleInRad)
                intRings = [
                    RotationLib.rotate2D(r.coords, angleInRad) for r in geom.interiors
                    ]
                return Polygon(extRing, intRings)

            elif GeomLib.isMultipart(geom):
                if isinstance(geom, MultiPoint):
                    return MultiPoint(RotationLib.rotate2D(geom.geoms, angleInRad))
                elif isinstance(geom, MultiLineString):
                    return MultiLineString(RotationLib.rotate2D(geom.geoms, angleInRad))
                elif isinstance(geom, MultiPolygon):
                    return MultiPolygon(RotationLib.rotate2D(geom.geoms, angleInRad))
                elif isinstance(geom, GeometryCollection):
                    return GeometryCollection(RotationLib.rotate2D(geom.geoms, angleInRad))

        return None



'''
Created on 10 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess
from t4gpd.commons.CaliperLib import CaliperLib


class MABR2(AbstractGeoprocess):
    '''
    classdocs
    '''

    @staticmethod
    def runWithArgs(row):
        rect, len1, len2 = CaliperLib().mabr(row.geometry)
        return { 'geometry': rect, 'mabr_len1': len1, 'mabr_len2': len2 }



'''
Created on 19 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import arctan2, pi 


class AngleLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def toDegrees(angle):
        return (180.0 * angle) / pi

    @staticmethod
    def toRadians(angle):
        return (pi * angle) / 180.0

    @staticmethod
    def normAzimuth(direction):
        result = arctan2(direction[1], direction[0])
        if (result < 0.0):
            return (2.0 * pi + result)
        else:
            return result

    @staticmethod
    def angleBetween(direction1, direction2):
        result = AngleLib.normAzimuth(direction2) - AngleLib.normAzimuth(direction1)
        if (result < 0.0):
            return (2.0 * pi + result)
        else:
            return result

    @staticmethod
    def angleBetweenNodes(node1, orig, node2):
        return AngleLib.angleBetween([ node1[0] - orig[0] , node1[1] - orig[1] ], [ node2[0] - orig[0] , node2[1] - orig[1] ])



'''
Created on 5 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import dtype
from sklearn.cluster import KMeans
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STKMeans(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, nclusters, verbose=True):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf

        if not isinstance(nclusters, int):
            raise IllegalArgumentTypeException(nclusters, 'int')
        self.kmeans = KMeans(n_clusters=nclusters, init='k-means++', n_init=10, max_iter=300)

        _fieldnames = []
        for _fieldname, _fieldtype in zip(self.inputGdf.columns, self.inputGdf.dtypes):
            if (_fieldtype == dtype(float)):
                _fieldnames.append(_fieldname)
        if (0 == len(_fieldnames)):
            raise Exception('There are no numeric (float) fields in the GeoDataFrame!')
        if verbose:
            print('The following fields are taken into account when processing KMeans: %s' % _fieldnames)
        self.data = self.inputGdf[_fieldnames]

    def run(self):
        _kmeans = self.kmeans.fit(self.data)

        result = GeoDataFrame.copy(self.inputGdf)
        result['kmeans_lbl'] = _kmeans.predict(self.data)
        return result



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import cos, pi, sin

from geopandas.geodataframe import GeoDataFrame
from geopandas.sindex import SpatialIndex
from rtree.index import Index 
from shapely.geometry import LineString, MultiLineString, Point
from shapely.ops import nearest_points, linemerge
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class RayCastingLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def prepareOrientedRays(masksGdf, masksSIdx, viewPoint):
        _, nearestPoint, _ = GeomLib.getNearestFeature(
            masksGdf, masksSIdx, viewPoint, buffDist=40.0)
        u = GeomLib.unitVector([viewPoint.x, viewPoint.y], [nearestPoint.x, nearestPoint.y])
        return [u, [-u[0], -u[1]]]

    @staticmethod
    def preparePanopticRays(nRays=64):
        angularOffset = (2.0 * pi) / nRays
        shootingDirs = [[float(cos(angularOffset * i)), float(sin(angularOffset * i))] for i in range(nRays)]
        return shootingDirs

    @staticmethod
    def singleRayCast2D(masksGdf, masksSIdx, viewPoint, shootingDir, rayLength=100.0):
        if not isinstance(masksGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(masksGdf, 'GeoDataFrame')
        if not isinstance(masksSIdx, (SpatialIndex, Index)):
            raise IllegalArgumentTypeException(masksSIdx, 'SpatialIndex or Index')
        if not isinstance(viewPoint, Point):
            raise IllegalArgumentTypeException(viewPoint, 'Point')

        # TL. 23.02.2021
        # To avoid: "Inconsistent coordinate dimensionality"
        viewPoint = Point((viewPoint.x, viewPoint.y))
        remotePoint = Point((viewPoint.x + shootingDir[0] * rayLength, viewPoint.y + shootingDir[1] * rayLength))
        ray = LineString([viewPoint, remotePoint])

        hitPoint, hitDist, hitMask = [None, rayLength, None]

        masksIds = list(masksSIdx.intersection(ray.bounds))
        for maskId in masksIds:
            mask = masksGdf.loc[maskId]
            maskGeom = mask.geometry
            if ray.crosses(maskGeom):
                _tmp = maskGeom.intersection(ray)
                _, rp = nearest_points(viewPoint, _tmp)
                dist = viewPoint.distance(rp)
                if (dist < hitDist):
                    hitPoint, hitDist, hitMask = rp, dist, mask

        if hitPoint is not None:
            ray = LineString([viewPoint, hitPoint])
        else:
            hitPoint = remotePoint

        return [ray, hitPoint, hitDist, hitMask]

    @staticmethod
    def singleRayCast25D(masksGdf, masksSIdx, viewPoint, shootingDir, rayLength,
                         elevationFieldName, background):
        if not isinstance(masksGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(masksGdf, 'GeoDataFrame')
        if not isinstance(masksSIdx, (SpatialIndex, Index)):
            raise IllegalArgumentTypeException(masksSIdx, 'SpatialIndex or Index')
        if not isinstance(viewPoint, Point):
            raise IllegalArgumentTypeException(viewPoint, 'Point')
        if elevationFieldName not in masksGdf:
            raise Exception('%s is not a relevant field name!' % (elevationFieldName))

        # TL. 23.02.2021
        # To avoid: "Inconsistent coordinate dimensionality"
        viewPoint = Point((viewPoint.x, viewPoint.y))
        remotePoint = Point((viewPoint.x + shootingDir[0] * rayLength, viewPoint.y + shootingDir[1] * rayLength))
        ray = LineString([viewPoint, remotePoint])

        hitPoint, hitDist, hitMask, hitHW = [None, rayLength, None, 0]

        masksIds = list(masksSIdx.intersection(ray.bounds))
        for maskId in masksIds:
            mask = masksGdf.loc[maskId]
            maskGeom = mask.geometry
            if ray.crosses(maskGeom):
                _tmp = maskGeom.intersection(ray)
                _, rp = nearest_points(viewPoint, _tmp)
                height, dist = mask[elevationFieldName], viewPoint.distance(rp)
                hw = height / dist
                if background and (hitHW < hw):
                    hitPoint, hitDist, hitMask, hitHW = rp, dist, mask, hw
                elif (not background) and (dist < hitDist):
                    hitPoint, hitDist, hitMask, hitHW = rp, dist, mask, hw

        if hitPoint is not None:
            ray = LineString([viewPoint, hitPoint])
        else:
            hitPoint = remotePoint

        return [ray, hitPoint, hitDist, hitMask, hitHW]

    @staticmethod
    def singleRayCastOnTopOfRoof(masksGdf, masksSIdx, viewPoint, shootingDir, rayLength,
                                 h0, elevationFieldName, background):
        
        if not isinstance(masksGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(masksGdf, 'GeoDataFrame')
        if not isinstance(masksSIdx, (SpatialIndex, Index)):
            raise IllegalArgumentTypeException(masksSIdx, 'SpatialIndex or Index')
        if not isinstance(viewPoint, Point):
            raise IllegalArgumentTypeException(viewPoint, 'Point')
        if elevationFieldName not in masksGdf:
            raise Exception('%s is not a relevant field name!' % (elevationFieldName))

        # TL. 23.02.2021
        # To avoid: "Inconsistent coordinate dimensionality"
        viewPoint = Point((viewPoint.x, viewPoint.y))
        remotePoint = Point((viewPoint.x + shootingDir[0] * rayLength, viewPoint.y + shootingDir[1] * rayLength))
        ray = LineString([viewPoint, remotePoint])

        hitPoint, hitDist, hitMask, hitHW = [None, rayLength, None, 0]

        masksIds = list(masksSIdx.intersection(ray.bounds))
        for maskId in masksIds:
            mask = masksGdf.loc[maskId]
            maskGeom = mask.geometry
            if ray.crosses(maskGeom):
                _tmp = maskGeom.intersection(ray)
                _, rp = nearest_points(viewPoint, _tmp)
                height, dist = mask[elevationFieldName], viewPoint.distance(rp)
                hw = (height - h0) / dist
                if background and (hitHW < hw):
                    hitPoint, hitDist, hitMask, hitHW = rp, dist, mask, hw
                elif (not background) and (dist < hitDist) and (0 < hw):
                    hitPoint, hitDist, hitMask, hitHW = rp, dist, mask, hw

        if hitPoint is not None:
            ray = LineString([viewPoint, hitPoint])
        else:
            hitPoint = remotePoint

        return [ray, hitPoint, hitDist, hitMask, hitHW]

    @staticmethod
    def multipleRayCast2D(masksGdf, masksSIdx, viewPoint, shootingDirs, rayLength=100.0):
        if not isinstance(masksGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(masksGdf, 'GeoDataFrame')
        if not isinstance(masksSIdx, (SpatialIndex, Index)):
            raise IllegalArgumentTypeException(masksSIdx, 'SpatialIndex or Index')
        if not isinstance(viewPoint, Point):
            raise IllegalArgumentTypeException(viewPoint, 'Point')

        rays, hitPoints, hitDists, hitMasks = [[], [], [], []]

        for shootingDir in shootingDirs:
            ray, hitPoint, hitDist, hitMask = RayCastingLib.singleRayCast2D(masksGdf, masksSIdx, viewPoint, shootingDir, rayLength)
            rays.append(ray)
            hitPoints.append(hitPoint)
            hitDists.append(hitDist)
            hitMasks.append(hitMask)

        return [MultiLineString(rays), hitPoints, hitDists, hitMasks]

    @staticmethod
    def multipleRayCast25D(masksGdf, masksSIdx, viewPoint, shootingDirs, rayLength,
                           elevationFieldName, background):
        if not isinstance(masksGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(masksGdf, 'GeoDataFrame')
        if not isinstance(masksSIdx, (SpatialIndex, Index)):
            raise IllegalArgumentTypeException(masksSIdx, 'SpatialIndex or Index')
        if not isinstance(viewPoint, Point):
            raise IllegalArgumentTypeException(viewPoint, 'Point')
        if elevationFieldName not in masksGdf:
            raise Exception('%s is not a relevant field name!' % (elevationFieldName))

        rays, hitPoints, hitDists, hitMasks, hitHWs = [[], [], [], [], []]

        for shootingDir in shootingDirs:
            ray, hitPoint, hitDist, hitMask, hitHW = RayCastingLib.singleRayCast25D(
                masksGdf, masksSIdx, viewPoint, shootingDir, rayLength,
                elevationFieldName, background)
            rays.append(ray)
            hitPoints.append(hitPoint)
            hitDists.append(hitDist)
            hitMasks.append(hitMask)
            hitHWs.append(hitHW)

        return [MultiLineString(rays), hitPoints, hitDists, hitMasks, hitHWs]

    @staticmethod
    def multipleRayCastOnTopOfRoof(masksGdf, masksSIdx, viewPoint, shootingDirs,
                                   rayLength, enclosingFeatures, elevationFieldName,
                                   background):
        if not isinstance(masksGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(masksGdf, 'GeoDataFrame')
        if not isinstance(masksSIdx, (SpatialIndex, Index)):
            raise IllegalArgumentTypeException(masksSIdx, 'SpatialIndex or Index')
        if not isinstance(viewPoint, Point):
            raise IllegalArgumentTypeException(viewPoint, 'Point')
        if elevationFieldName not in masksGdf:
            raise Exception('%s is not a relevant field name!' % (elevationFieldName))

        rays, hitPoints, hitDists, hitMasks, hitHWs = [[], [], [], [], []]
        h0 = min([f[elevationFieldName] for f in enclosingFeatures])

        for shootingDir in shootingDirs:
            ray, hitPoint, hitDist, hitMask, hitHW = RayCastingLib.singleRayCastOnTopOfRoof(
                masksGdf, masksSIdx, viewPoint, shootingDir, rayLength, h0,
                elevationFieldName, background)
            rays.append(ray)
            hitPoints.append(hitPoint)
            hitDists.append(hitDist)
            hitMasks.append(hitMask)
            hitHWs.append(hitHW)

        return [MultiLineString(rays), hitPoints, hitDists, hitMasks, hitHWs]

    @staticmethod
    def assessStreetWidth(masksGdf, masksSIdx, viewPoint, rayLength=100.0):
        if GeomLib.isAnIndoorPoint(viewPoint, masksGdf, masksSIdx):
            return None, -1, [-1, -1]

        _shootingDirs = RayCastingLib.prepareOrientedRays(
            masksGdf, masksSIdx, viewPoint)
        rays, _, hitDists, _ = RayCastingLib.multipleRayCast2D(
            masksGdf, masksSIdx, viewPoint, _shootingDirs, rayLength)
        return linemerge(rays), sum(hitDists), hitDists



'''
Created on 24 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame

from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STSnappingPointsOnLines(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, pointsGdf, linesGdf, stepCountFieldname=None):
        '''
        Constructor
        '''
        if not isinstance(pointsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(pointsGdf, 'GeoDataFrame')
        self.pointsGdf = pointsGdf

        if not isinstance(linesGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(linesGdf, 'GeoDataFrame')
        self.linesGdf = linesGdf
        self.spatialIdx = linesGdf.sindex

        if stepCountFieldname is not None:
            if stepCountFieldname not in pointsGdf:
                raise Exception('%s is not a relevant field name!' % (stepCountFieldname))
        self.stepCountFieldname = stepCountFieldname

    def __check(self, checksum):
        prevStepCount, prevCurvAbsc = None, None
        for stepCount in sorted(checksum.keys()):
            if (prevStepCount is not None) and (prevCurvAbsc > checksum[stepCount]):
                print('Warning: curvilinear abscissa inconsistencies between step %d (%.2f m) and %d (%.2f m)' % 
                      (prevStepCount, prevCurvAbsc, stepCount, checksum[stepCount]))
            prevStepCount, prevCurvAbsc = stepCount, checksum[stepCount]

    def run(self):
        checksum, rows = dict(), []
        for _, rowPoint in self.pointsGdf.iterrows():
            geomPoint = rowPoint.geometry
            buffDist = 40.0
            minDist, nearestPoint, nearestRow = GeomLib.getNearestFeature(
                self.linesGdf, self.spatialIdx, geomPoint, buffDist)

            nearestLine = nearestRow.geometry
            curvilinearAbscissa = nearestLine.project(nearestPoint)
            if self.stepCountFieldname is not None:
                checksum[rowPoint[self.stepCountFieldname]] = curvilinearAbscissa

            rows.append(self.updateOrAppend(rowPoint, {
                # 'geometry': LineString((geomPoint, nearestPoint)),
                'geometry': nearestPoint,
                'dist_to_l': minDist,
                'curv_absc': curvilinearAbscissa 
                }))

        if self.stepCountFieldname is not None:
            self.__check(checksum)
        return GeoDataFrame(rows, crs=self.pointsGdf.crs)



'''
Created on 31 janv. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import numbers

from geopandas import GeoDataFrame
from shapely.affinity import rotate
from shapely.geometry import Polygon
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.future.STCrossroadsGeneration import STCrossroadsGeneration

import matplotlib.pyplot as plt


class MultipleMarkerStyles(GeoProcess):
    '''
    classdocs
    '''
    MPATH = '__mpath__'

    def __init__(self, nodes, patterns=None, left_on='recId_fwt', right_on='gid',
                 rotation=None, basemap=None, marker_color='black', marker_size=2000,
                 alpha=None, nbranchs=8):
        '''
        Constructor
        '''
        if not isinstance(nodes, GeoDataFrame):
            raise IllegalArgumentTypeException(nodes, 'GeoDataFrame')
        if left_on not in nodes:
            raise IllegalArgumentTypeException(left_on, 'is not a valid field name')
        self.left_on = left_on

        if patterns is None:
            _patterns = STCrossroadsGeneration(nbranchs, length=100.0, width=10.0,
                                               mirror=False, withBranchs=True, withSectors=True,
                                               crs='epsg:2154', magnitude=0).run()
        else:
            if not isinstance(patterns, GeoDataFrame):
                raise IllegalArgumentTypeException(patterns, 'GeoDataFrame')
            _patterns = patterns.copy()
        if right_on not in _patterns:
            raise IllegalArgumentTypeException(right_on, 'is not a valid field name')

        if isinstance(rotation, numbers.Number):
            _patterns[MultipleMarkerStyles.MPATH] = _patterns.geometry.apply(
                lambda g: rotate(g, rotation, origin='center').exterior.coords)
            self.rotation = None
        else:
            _patterns[MultipleMarkerStyles.MPATH] = _patterns.geometry.apply(
                lambda g: g.exterior.coords)
            if isinstance(rotation, str) and (rotation not in nodes):
                raise IllegalArgumentTypeException(rotation, 'is not a valid field name')
            self.rotation = rotation
        self.mapOfPatterns = _patterns[[right_on, MultipleMarkerStyles.MPATH]].set_index(right_on).to_dict('index')

        self.nodes = nodes.copy()
        self.nodes = self.nodes.merge(_patterns[[right_on, MultipleMarkerStyles.MPATH]], how='left',
                                      left_on=left_on, right_on=right_on)

        self.basemap = plt if basemap is None else basemap
        self.marker_color = marker_color
        self.marker_size = marker_size
        self.alpha = alpha

    def run(self):
        if self.rotation is None:
            for _pattId in self.mapOfPatterns.keys():
                _nodes = self.nodes[ self.nodes[self.left_on] == _pattId ]
                _mpath = self.mapOfPatterns[_pattId][MultipleMarkerStyles.MPATH]
                self.basemap.scatter(_nodes.centroid.x, _nodes.centroid.y,
                                     c=self.marker_color, s=self.marker_size,
                                     alpha=self.alpha, marker=_mpath)
        else:
            for _pattId in self.mapOfPatterns.keys():
                _mpath = Polygon(self.mapOfPatterns[_pattId][MultipleMarkerStyles.MPATH])
                _nodes = self.nodes[ self.nodes[self.left_on] == _pattId ]
                for _, _row in _nodes.iterrows():
                    _rot, _geom = _row[self.rotation], _row.geometry.centroid
                    _curr_mpath = rotate(_mpath, 360 - _rot, origin='center').exterior.coords
                    self.basemap.scatter(_geom.x, _geom.y,
                                         c=self.marker_color, s=self.marker_size,
                                         alpha=self.alpha, marker=_curr_mpath)



'''
Created on 31 aug. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import Point, Polygon
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class VTUWriter(GeoProcess):
    '''
    classdocs
    
    For more information, please see
    https://www.vtk.org/wp-content/uploads/2015/04/file-formats.pdf
    '''
    VTKTYPE = 7  # VTK_POLYGON=7, VTK_POLY_LINE=4, VTK_POLY_VERTEX=2

    def __init__(self, inputGdf, outputFile):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf

        self.cellData = dict()
        for fname in self.inputGdf.columns:
            ftype = str(self.inputGdf.dtypes[fname])
            if 'int64' == ftype:
                self.cellData[fname] = { 'type':'Int32', 'values':[] }
            elif 'float64' == ftype:
                self.cellData[fname] = { 'type':'Float32', 'values':[] }
            # LE CHARGEMENT DE CHAMPS DE TYPES STRING SEMBLE PLANTER PARAVIEW
            # elif (object == ftype):
                # self.cellData[fname] = { 'type':'String', 'values':[] }

        self.outputFile = outputFile

    def __addAndGetANode(self, x, y, z=0.0):
        hashCoord = GeomLib.hashCoords(x, y, z)
        if hashCoord in self.ciVertices:
            nodeIndex = self.ciVertices[hashCoord]
        else:
            nodeIndex = len(self.ciVertices)
            self.ciVertices[hashCoord] = nodeIndex
            self.icVertices[nodeIndex] = [x, y, z]
        return nodeIndex           

    def __addContour(self, geom):
        contour = []
        if isinstance(geom, Polygon):
            if (GeomLib.isHoled(geom)):
                print("*** Holed polygon:: %s" % geom.wkt)
            elif not GeomLib.isConvex(geom):
                print("*** Concave polygon:: %s" % geom.wkt)

            for xyz in geom.exterior.coords:
                p = Point(xyz)
                if p.has_z:
                    nodeIndex = self.__addAndGetANode(p.x, p.y, p.z)
                else:
                    nodeIndex = self.__addAndGetANode(p.x, p.y)
                contour.append(nodeIndex)

        return contour
        # Remove final point
        # return contour[:-1]

    def run(self):
        self.ciVertices = dict()
        self.icVertices = dict()
        contours, offsets = [], []

        for _, row in self.inputGdf.iterrows():
            for fieldname in self.cellData:
                self.cellData[fieldname]['values'].append(row[fieldname])

            geom = row.geometry
            if GeomLib.isMultipart(geom):
                for g in geom.geoms:
                    ctr = self.__addContour(g)
                    contours.append(ctr)

                    if (0 == len(offsets)):
                        offsets.append(len(ctr))
                    else:
                        offsets.append(offsets[-1] + len(ctr))
            else:
                ctr = self.__addContour(geom)
                contours.append(ctr)

                if (0 == len(offsets)):
                    offsets.append(len(ctr))
                else:
                    offsets.append(offsets[-1] + len(ctr))

        with open(self.outputFile, 'w') as f: 
            f.write("<VTKFile type='UnstructuredGrid' version='0.1'>\n")
            f.write("\t<UnstructuredGrid>\n")
            f.write("\t\t<Piece NumberOfPoints='%d' NumberOfCells='%d'>\n" % (len(self.icVertices), len(contours)))

            # ===== CELLDATA =====
            f.write("\t\t\t<CellData Scalars='scalars'>\n")
            f.writelines([
                "\t\t\t\t<DataArray Name='%s' type='%s' format='ascii'>\n%s\t\t\t\t</DataArray>\n" % 
                    (k, v['type'], ' '.join(str(i) for i in v['values']))
                for k, v in list(self.cellData.items())
            ])
            '''
            f.writelines(
                map(lambda(k, v): 
                    "\t\t\t\t<DataArray Name='%s' type='%s' format='ascii'>\n%s\t\t\t\t</DataArray>\n" % 
                    (k, v['type'], ' '.join(str(i) for i in v['values'])), self.cellData.iteritems())
                )
            '''
            f.write("\t\t\t</CellData>\n")

            f.write("\t\t\t<Points>\n")
            f.write("\t\t\t\t<DataArray type='Float32' NumberOfComponents='3' format='ascii'>\n")
            f.writelines([
                '%f %f %f\n' % (v[0], v[1], v[2]) for k, v in sorted(self.icVertices.items())
            ])
            # f.writelines(map(lambda(k, v): '%f %f %f\n' % (v[0], v[1], v[2]), sorted(self.icVertices.iteritems())))  # COORDS
            f.write("\t\t\t\t</DataArray>\n")
            f.write("\t\t\t</Points>\n")
            f.write("\t\t\t<Cells>\n")
            f.write("\t\t\t\t<DataArray type='Int32' Name='connectivity' format='ascii'>\n")
            f.writelines([' '.join(str(i) for i in ctr) + ' ' for ctr in contours] + ['\n'])  # CONNECTIVITY
            f.write("\t\t\t\t</DataArray>\n")
            f.write("\t\t\t\t<DataArray type='Int32' Name='offsets' format='ascii'>\n")
            f.writelines(['%d ' % i for i in offsets] + ['\n'])  # OFFSETS
            f.write("\t\t\t\t</DataArray>\n")
            f.write("\t\t\t\t<DataArray type='Int32' Name='types' format='ascii'>\n")
            f.writelines(['%d ' % (VTUWriter.VTKTYPE) for _ in range(len(contours))] + ['\n'])  # TYPES
            f.write("\t\t\t\t</DataArray>\n")
            f.write("\t\t\t</Cells>\n")
            f.write("\t\t</Piece>\n")
            f.write("\t</UnstructuredGrid>\n")
            f.write("</VTKFile>\n")

            print('%s has been written!' % self.outputFile)



'''
Created on 16 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from inspect import isclass

from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class STGeoProcess(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, geoprocessToApply, inputGdf):
        '''
        Constructor
        '''
        if not (self.__is_a_geoprocess(geoprocessToApply) or 
                self.__is_a_collection_of_geoprocesses(geoprocessToApply)):
            raise IllegalArgumentTypeException(geoprocessToApply, '(list of) AbstractGeoprocess')

        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')

        if (self.__is_a_geoprocess(geoprocessToApply)):
            self.geoprocessToApply = [geoprocessToApply]
        else:
            self.geoprocessToApply = geoprocessToApply
        self.inputGdf = inputGdf

    def __is_a_geoprocess(self, obj):
        return (isinstance(obj, AbstractGeoprocess) or 
                (isclass(obj) and issubclass(obj, AbstractGeoprocess)))

    def __is_a_collection_of_geoprocesses(self, obj):
        return (isinstance(obj, (list, tuple)) and all([self.__is_a_geoprocess(_obj) for _obj in obj]))

    def run(self):
        rows = []
        for _, row in self.inputGdf.iterrows():
            result = dict()
            for op in self.geoprocessToApply:
                result.update(op.runWithArgs(row))
            rows.append(self.updateOrAppend(row, result))
        return GeoDataFrame(rows, crs=self.inputGdf.crs)

    '''
    def _run(self, inputGdf):
        rows = []
        for _, row in inputGdf.iterrows():
            result = dict()
            for op in self.geoprocessToApply:
                result.update(op.runWithArgs(row))
            rows.append(self.updateOrAppend(row, result))
        return GeoDataFrame(rows, crs=inputGdf.crs)

    def parallel(self):
        ncpu, nrows = cpu_count(), len(self.inputGdf)
        loi = unique(linspace(0, nrows, ncpu).astype(int))
        gdfs = [self.inputGdf.iloc[loi[i - 1]:loi[i]] for i in range(1, len(loi))]
        with Pool(processes=ncpu) as pool:
            gdfs = pool.map(self._run, gdfs)
        return concat(gdfs)
    '''



'''
Created on 21 nov. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from dijkstar.algorithm import find_path, NoPathError
from shapely.geometry import MultiLineString
from shapely.wkt import loads

from t4gpd.commons.graph.AbstractUrbanGraphLib import AbstractUrbanGraphLib


class ShortestPathLib(AbstractUrbanGraphLib):
    '''
    classdocs
    '''

    def __init__(self, roads, roi=None):
        '''
        Constructor
        '''
        self.roads, self.graph, self.ciVertices, self.icVertices, self.edges = AbstractUrbanGraphLib.initializeGraph(roads, roi)
        self.spatialIdx = self.roads.sindex
        self.roi = roi

    def shortestPath(self, fromPoint, toPoint):
        # Backup the input graph
        _graph, _ciVertices, _icVertices, _edges = AbstractUrbanGraphLib.backupUrbanGraph(
            self.graph, self.ciVertices, self.icVertices, self.edges)

        if (fromPoint == toPoint):
            return MultiLineString([]), 0.0

        str_coord = AbstractUrbanGraphLib.hashCoord([fromPoint.x, fromPoint.y])
        if str_coord in self.ciVertices:
            fromIdx = self.ciVertices[str_coord]
        else:
            fromIdx, _ = AbstractUrbanGraphLib.addEndPoint(
                self.roads, self.spatialIdx, self.graph, self.ciVertices, self.icVertices, self.edges, fromPoint)

        str_coord = AbstractUrbanGraphLib.hashCoord([toPoint.x, toPoint.y])
        if str_coord in self.ciVertices:
            toIdx = self.ciVertices[str_coord]
        else:
            toIdx, _ = AbstractUrbanGraphLib.addEndPoint(
                self.roads, self.spatialIdx, self.graph, self.ciVertices, self.icVertices, self.edges, toPoint)

        try:
            tmp = find_path(self.graph, fromIdx, toIdx, cost_func=AbstractUrbanGraphLib.cost_function)
            pathGeom = MultiLineString([loads(wkt) for _, wkt in tmp.edges])
            pathLen = tmp.total_cost

        except NoPathError as error:
            pathGeom, pathLen = None, None
            print(error)

        finally:
            # Rollback to the input graph (must be executed under all circumstances)
            self.graph, self.ciVertices, self.icVertices, self.edges = _graph, _ciVertices, _icVertices, _edges

        return pathGeom, pathLen



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from shapely.geometry import LinearRing, LineString, Point, Polygon

from t4gpd.commons.GeomLib import GeomLib


class UrbanGraphLibOld(object):
    '''
    classdocs
    '''

    def __init__(self):
        '''
        Constructor
        '''
        self.ciVertices = dict()
        self.icVertices = dict()
        self.edges = dict()

    def add(self, features):
        if isinstance(features, list):
            for feature in features:
                self.add(feature)
        else:
            if GeomLib.isAShapelyGeometry(features):
                geom = features
            else:
                geom = features.geometry

            if GeomLib.isMultipart(geom):
                for g in geom.geoms:
                    self.add(g)
            elif isinstance(geom, LineString):
                self.__addPolyline(geom)
            elif isinstance(geom, Polygon):
                self.__addPolyline(geom.exterior)
                for g in geom.interiors:
                    self.__addPolyline(g)
            else:
                raise Exception('Illegal argument in UrbanGraphLib.add(...)!')
        # print('self.icVertices: %s' % self.icVertices)
        # print('self.edges: %s' % self.edges)

    def __addPolyline(self, line):
        if isinstance(line, LinearRing):
            coords = line.coords[:-1]
        else:
            coords = line.coords

        if (2 <= len(coords)):
            prev = self.__addAndGetANode(coords[0])
            for i in range(1, len(coords)):
                curr = self.__addAndGetANode(coords[i]);
                if self.edges.get(prev) is None:
                    self.edges[prev] = set([ curr ])
                else:
                    self.edges[prev].add(curr)
                if self.edges.get(curr) is None:
                    self.edges[curr] = set([ prev ])
                else:
                    self.edges[curr].add(prev)
                prev = curr;

    @staticmethod
    def hashCoord(coord):
        str_coord = ('%f_%f') % (coord[0], coord[1])
        return str_coord

    def __addAndGetANode(self, coord):
        str_coord = UrbanGraphLibOld.hashCoord(coord)
        if self.ciVertices.get(str_coord) is None:
            nodeIndex = len(self.ciVertices)
            self.ciVertices[str_coord] = nodeIndex
            self.icVertices[nodeIndex] = coord
            return nodeIndex   
        else:
            nodeIndex = self.ciVertices[str_coord]
            return nodeIndex   

    def getUniqueRoadsSectionsNodes(self):
        result = list()
        for k, v in list(self.edges.items()):
            nbConnectedVertices = len(v)
            if 2 != nbConnectedVertices:
                result.append({ 'geometry': Point(self.icVertices[k]), 'gid': k,
                               'valency': nbConnectedVertices })
        return result

    def __getNextVertex(self, prev, curr):
        succ = list(self.edges[curr])
        if 2 == len(succ):
            if (prev == succ[0]) and not (prev == succ[1]):
                return succ[1]
            elif (prev == succ[1]) and not (prev == succ[0]):
                return succ[0]
        return None

    def __burnEdge(self, idx1, idx2, burnedEdges):
        minIdx = min(idx1, idx2)
        maxidx = max(idx1, idx2)
        if minIdx in burnedEdges:
            burnedEdges[minIdx].add(maxidx)
        else:
            burnedEdges[minIdx] = set({ maxidx })

    def _buildRoadSection(self, vertexIndex, nextVertexIndex, burnedEdges):
        result = []

        prevIdx = vertexIndex
        result.append(self.icVertices[prevIdx])
        currIdx = nextVertexIndex
        result.append(self.icVertices[currIdx])
        self.__burnEdge(prevIdx, currIdx, burnedEdges)

        nextIdx = self.__getNextVertex(prevIdx, currIdx)
        while nextIdx is not None:
            result.append(self.icVertices[nextIdx])
            prevIdx = currIdx
            currIdx = nextIdx
            self.__burnEdge(prevIdx, currIdx, burnedEdges)

            if (nextIdx == vertexIndex):
                # current road section is a loop, stop the process
                nextIdx = None
            else:
                nextIdx = self.__getNextVertex(prevIdx, currIdx)            

        return LineString(result)

    def getUniqueRoadsSections(self):
        result = []
        burnedEdges = dict()
        for vertexIndex in list(self.edges.keys()):
            nbConnectedVertices = len(self.edges[vertexIndex])
            if 2 != nbConnectedVertices:
                # current vertex is a 'boundary' of a 'road section'
                for nextVertexIndex in self.edges[vertexIndex]:
                    minIdx = min(vertexIndex, nextVertexIndex)
                    maxIdx = max(vertexIndex, nextVertexIndex)
                    if not (minIdx in burnedEdges and (maxIdx in burnedEdges[minIdx])):
                        geom = self._buildRoadSection(vertexIndex, nextVertexIndex, burnedEdges)
                        result.append({ 'geometry': geom, 'gid': len(result),
                                       'distance': geom.length })
        return result

    def getUniqueRoadsSectionsWithoutCulDeSac(self):
        result = []

        rs = self.getUniqueRoadsSections()
        for row in rs:
            _coords = list(row['geometry'].coords)
            startHash = UrbanGraphLibOld.hashCoord(_coords[0])
            stopHash = UrbanGraphLibOld.hashCoord(_coords[-1])
            if ((1 < len(self.edges[self.ciVertices[startHash]])) and
                (1 < len(self.edges[self.ciVertices[stopHash]]))):
                result.append(row)
        return result



'''
Created on 11 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from timeit import default_timer as timer


class GeoProcess(object):
    '''
    classdocs
    '''

    def __elapsedTime(self, start, stop):
        delta = stop - start
        print('Elapsed time: %.3f s (%02d:%02d:%02d)' % (
            delta, int(delta) / 3600, (int(delta) % 3600) / 60, (int(delta) % 3600) % 60))

    def execute(self):
        start = timer()
        result = self.run()
        stop = timer()
        self.__elapsedTime(start, stop)
        return result

    def executeWithArgs(self, args):
        start = timer()
        result = self.runWithArgs(args)
        stop = timer()
        self.__elapsedTime(start, stop)
        return result

    def run(self):
        raise NotImplementedError('subclasses must override run()!')

    def runWithArgs(self, args):
        raise NotImplementedError('subclasses must override runWithArgs()!')

    def updateOrAppend(self, inSeries, addDict):
        result = dict(inSeries)
        result.update(addDict)
        return result

    def appendNewItems(self, inSeries, addDict):
        result = dict(inSeries)
        for k, v in addDict.items():
            if k not in result:
                result[k] = v
        return result



'''
Created on 15 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from functools import reduce

from numpy import cos, sin, sqrt, pi
from pandas.core.common import flatten
from shapely.geometry import GeometryCollection, LinearRing, LineString, MultiLineString, MultiPoint, MultiPolygon, Point, Polygon
from shapely.ops import nearest_points, transform, unary_union
from t4gpd.commons.AngleLib import AngleLib
from t4gpd.commons.Epsilon import Epsilon
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.ListUtilities import ListUtilities


class GeomLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def addZCoordinateIfNecessary(obj):
        if GeomLib.isAShapelyGeometry(obj):
            if GeomLib.is3D(obj):
                return obj
            return transform(lambda x, y, z=None: (x, y, 0.0), obj)
        raise IllegalArgumentTypeException(obj, 'Shapely geometry')

    @staticmethod
    def areCollinear(u, v, epsilon=Epsilon.EPSILON):
        return Epsilon.isZero(GeomLib.crossProduct(u, v), epsilon)

    @staticmethod
    def areAligned(a, b, c, epsilon=Epsilon.EPSILON):
        ab = GeomLib.vector_to(a, b)
        ac = GeomLib.vector_to(a, c)
        return GeomLib.areCollinear(ab, ac, epsilon)

    @staticmethod
    def crossProduct(u, v):
        return [0.0, 0.0, u[0] * v[1] - u[1] * v[0]]

    @staticmethod
    def cutsLineStringByCurvilinearDistance(geom, curvDistance):
        if isinstance(geom, LineString):
            if (0 < curvDistance < geom.length):
                # The following instruction is not robust:
                # result = split(geom, geom.interpolate(curvDistance))
                # return list(result.geoms)
                coords = list(geom.coords)
                for i, p in enumerate(coords):
                    pdist = geom.project(Point(p))
                    if (curvDistance == pdist):
                        return [ LineString(coords[:i + 1]), LineString(coords[i:]) ]
                    elif (curvDistance < pdist):
                        cp = geom.interpolate(curvDistance)
                        cp = (cp.x, cp.y, cp.z) if geom.has_z else (cp.x, cp.y)
                        return [
                            LineString(coords[:i] + [cp]),
                            LineString([cp] + coords[i:]) ]
            return [ geom ]
        raise IllegalArgumentTypeException(geom, 'LineString')

    @staticmethod
    def distFromTo(a, b):
        dx = b[0] - a[0]
        dy = b[1] - a[1]
        return sqrt(dx * dx + dy * dy)

    @staticmethod
    def dotProduct(u, v):
        return u[0] * v[0] + u[1] * v[1]

    @staticmethod
    def extractFeaturesWithin(zone, inputGdf, inputSpatialIndex):
        ids = list(inputSpatialIndex.intersection(zone.bounds))

        features = []
        for _id in ids:
            row = dict(inputGdf.loc[_id])
            rowGeom = row['geometry']
            if zone.intersects(rowGeom):
                # Use a buffer to avoid slivers
                row['geometry'] = rowGeom.intersection(zone).buffer(0.001)
                features.append(row)

        return features

    @staticmethod
    def extractGeometriesWithin(zone, inputGdf, inputSpatialIndex):
        features = GeomLib.extractFeaturesWithin(zone, inputGdf, inputSpatialIndex)
        geoms = [feature['geometry'] for feature in features]

        if (0 < len(geoms)):
            result = unary_union(geoms)
            # Use a buffer to avoid slivers
            return result.buffer(0.001)
        return Polygon()

    @staticmethod
    def flattenGeometry(obj):
        _develop = lambda geom: [_develop(g) for g in geom.geoms] if GeomLib.isMultipart(geom) else [geom]

        if GeomLib.isAShapelyGeometry(obj):
            return list(flatten(_develop(obj)))

        return None

    @staticmethod
    def forceZCoordinateToZ0(obj, z0=0.0):
        if GeomLib.isAShapelyGeometry(obj):
            return transform(lambda x, y, z=None: (x, y, z0), obj)
        raise IllegalArgumentTypeException(obj, 'Shapely geometry')

    @staticmethod
    def fromMultiLineStringToLengths(multiline):
        if not isinstance(multiline, MultiLineString):
            raise IllegalArgumentTypeException(multiline, 'MultiLineString')
            # return []
        return [line.length for line in multiline.geoms]

    @staticmethod
    def fromRayLengthsToPolygon(rayLengths, origin=Point((0.0, 0.0))):
        nRays = len(rayLengths)
        offset = 2 * pi / nRays
        return Polygon([(origin.x + rayLengths[i] * cos(i * offset),
                         origin.y + rayLengths[i] * sin(i * offset)
                         ) for i in range(nRays)])

    @staticmethod
    def fromPolygonToListOfTriangles(polygon):
        if not isinstance(polygon, Polygon):
            raise IllegalArgumentTypeException(polygon, 'Polygon')
        if GeomLib.isHoled(polygon):
            raise Exception('The processing of polygons with holes remains to be developed!')

        p = list(GeomLib.normalizeRingOrientation(polygon).exterior.coords)[:-1]
        i, n, result = 0, len(p), []
        while n > 0:
            if (3 >= n):
                result.append(Polygon(p))
                n, p = 0, []
            else:
                i, ip1, ip2 = i % n, (i + 1) % n, (i + 2) % n

                if ((0 <= GeomLib.zCrossProduct(p[i], p[ip1], p[ip2])) and
                    LineString([p[i], p[ip2]]).within(Polygon(p))):
                    result.append(Polygon([ p[i], p[ip1], p[ip2] ]))
                    del(p[ip1])
                    n = n - 1

                i += 1

        return result 
        
    @staticmethod
    def getEnclosingFeatures(inputGdf, inputSpatialIndex, point):
        result = []
        ids = list(inputSpatialIndex.intersection(point.bounds))
        for _id in ids:
            _feature = inputGdf.loc[_id]
            _geom = _feature.geometry
            if point.within(_geom):
                result.append(_feature)
        return result        

    @staticmethod
    def getCircumcircle(a, b, c):
        if GeomLib.areAligned(a, b, c):
            ab = GeomLib.distFromTo(a, b)
            bc = GeomLib.distFromTo(b, c)
            ca = GeomLib.distFromTo(c, a)
            if Epsilon.EPSILON < min([ab, bc, ca]):
                return None
            diam = max([ab, bc, ca])
            if diam == ab:
                return [GeomLib.getMidPoint(a, b), diam / 2.0]
            elif diam == bc:
                return [GeomLib.getMidPoint(b, c), diam / 2.0]
            elif diam == ca:
                return [GeomLib.getMidPoint(c, a), diam / 2.0]
        else:
            line1 = GeomLib.getLineSegmentBisector(a, b)
            line2 = GeomLib.getLineSegmentBisector(b, c)
            center = GeomLib.intersect_line_line(line1, line2)
            return [center, GeomLib.distFromTo(a, center)]

    @staticmethod
    def getInteriorPoint(geom):
        if isinstance(geom, Point):
            return geom
        elif isinstance(geom, Polygon):
            centroid = geom.centroid
            if centroid.within(geom):
                return centroid
            extPts = geom.exterior.coords
            dists = [{'dist': centroid.distance(Point(v)), 'v': Point(v)} for v in extPts]
            # dists = sorted(dists, key=lambda l: l['dist'], reverse=True)
            dists = sorted(dists, key=lambda l: l['dist'], reverse=False)

            for q in [dist['v'] for dist in dists]:
                p = Point((centroid.x + q.x) / 2.0, (centroid.y + q.y) / 2.0)
                while not p.within(geom):
                    if (0.5 > p.distance(q)):
                        break
                    p = Point((p.x + q.x) / 2.0, (p.y + q.y) / 2.0)
                if p.within(geom):
                    return p
            return dists[0]['v']
        raise Exception('GeomLib.getInteriorPoint(...): implementation to be continued!')
    
    @staticmethod
    def getLineSegmentBisector(a, b):
        # The resulting line is represented by a point and a vector
        ab = GeomLib.vector_to(a, b)
        c = GeomLib.getMidPoint(a, b)
        return [ c, [-ab[1], ab[0]] ]

    @staticmethod
    def getLineStringOrientation(line):
        if not isinstance(line, LineString):
            raise IllegalArgumentTypeException(line, 'LineString')

        coords = list(line.coords)
        a, b = Point(coords[0]), Point(coords[-1])
        meanDir = (b.x - a.x, b.y - a.y)
        azimuth = AngleLib.normAzimuth(meanDir)
        if azimuth > pi:
            azimuth -= pi
        return AngleLib.toDegrees(azimuth)

    @staticmethod
    def getListOfShapelyPoints(obj, withoutClosingLoops=False):
        if isinstance(obj, Point):
            return [obj]
        elif isinstance(obj, LineString):
            if withoutClosingLoops and (obj.is_closed):
                return [Point(p) for p in obj.coords[:-1]]
            return [Point(p) for p in obj.coords]
        elif isinstance(obj, Polygon):
            result = GeomLib.getListOfShapelyPoints(obj.exterior, withoutClosingLoops)
            for ring in obj.interiors:
                result += GeomLib.getListOfShapelyPoints(ring, withoutClosingLoops)
            return result
        elif GeomLib.isMultipart(obj):
            return reduce(lambda a, b: a + b, [GeomLib.getListOfShapelyPoints(g, withoutClosingLoops) for g in obj.geoms])            
        raise IllegalArgumentTypeException(obj, 'Shapely geometry')

    @staticmethod
    def getMidPoint(a, b):
        return [ (a[0] + b[0]) / 2.0, (a[1] + b[1]) / 2.0 ]

    @staticmethod
    def getNearestFeature(inputGdf, inputSpatialIndex, point, buffDist, incScale=sqrt(2)):
        if not isinstance(point, Point):
            raise IllegalArgumentTypeException(point, 'Point')

        minDist, nearestPoint, nearestRow = float('inf'), None, None

        ids = []
        while (0 == len(ids)):
            ids = list(inputSpatialIndex.intersection(point.buffer(buffDist, -1).bounds))
            buffDist *= incScale

        for _id in ids:
            row = inputGdf.loc[_id]
            rowGeom = row.geometry
            dist = point.distance(rowGeom)
            if (dist < minDist):
                minDist = dist
                _, nearestPoint = nearest_points(point, rowGeom)
                nearestRow = row

        return [minDist, nearestPoint, nearestRow]

    @staticmethod
    def getStraightLineEquation(line):
        # The input line is represented by a point and a vector
        # The output result is [a, b, c] 
        # where a * x + b * y + c = 0
        point, vector = line
        a = vector[1]
        b = -vector[0]
        c = -a * point[0] - b * point[1]
        return [a, b, c]

    @staticmethod
    def hashCoords(x, y, z=0):
        _hash = 17
        _hash = 37 * _hash + hash(x)
        _hash = 37 * _hash + hash(y)
        _hash = 37 * _hash + hash(z)
        return _hash

    @staticmethod
    def intersect_line_line(line1, line2):
        # The input lines are both represented by a point and a vector
        a1, b1, c1 = GeomLib.getStraightLineEquation(line1)
        a2, b2, c2 = GeomLib.getStraightLineEquation(line2)
        det = a1 * b2 - a2 * b1
        if (0.0 == det):
            return None
        return [(b1 * c2 - b2 * c1) / det, (a2 * c1 - a1 * c2) / det]

    @staticmethod
    def is3D(obj):
        return GeomLib.isAShapelyGeometry(obj) and obj.has_z

    @staticmethod
    def getAnchoringBuildingId(point, buildings, spatialIndex):
        # isAnAnchoringPolygon = isABorderPoint OR isAnInsidePoint
        isAnAnchoringPolygon = lambda point, polygon: (
            point.relate(polygon) in ['F0FFFF212', '0FFFFF212'])
        buildingsIds = list(spatialIndex.intersection(point.bounds))
        for buildingId in buildingsIds:
            buildingGeom = buildings.loc[buildingId].geometry
            if isAnAnchoringPolygon(point, buildingGeom):
                return buildingId
        return None

    @staticmethod
    def isABorderPoint(point, buildings, spatialIndex):
        isAPointOnTheBorder = lambda point, polygon: ('F0FFFF212' == point.relate(polygon))
        buildingsIds = list(spatialIndex.intersection(point.bounds))
        for buildingId in buildingsIds:
            buildingGeom = buildings.loc[buildingId].geometry
            if isAPointOnTheBorder(point, buildingGeom):
                return True
        return False

    @staticmethod
    def isAnIndoorPoint(point, buildings, spatialIndex):
        isAnInsidePoint = lambda point, polygon: ('0FFFFF212' == point.relate(polygon))
        buildingsIds = list(spatialIndex.intersection(point.bounds))
        for buildingId in buildingsIds:
            buildingGeom = buildings.loc[buildingId].geometry
            if isAnInsidePoint(point, buildingGeom):
                return True
        return False

    @staticmethod
    def isAnOutdoorPoint(point, buildings, spatialIndex):
        isAnOutsidePoint = lambda point, polygon: ('FF0FFF212' == point.relate(polygon))
        buildingsIds = list(spatialIndex.intersection(point.bounds))
        for buildingId in buildingsIds:
            buildingGeom = buildings.loc[buildingId].geometry
            if not isAnOutsidePoint(point, buildingGeom):
                return False
        return True

    @staticmethod
    def isAShapelyGeometry(obj):
        return isinstance(obj, (GeometryCollection, LineString, MultiLineString,
                                MultiPoint, MultiPolygon, Point, Polygon))

    @staticmethod
    def isCCW(ring):
        return (0 < GeomLib.ringSignedArea(ring))

    @staticmethod
    def isConvex(obj):
        if isinstance(obj, Polygon) and not GeomLib.isHoled(obj):
            coords = list(obj.exterior.coords)
            if 4 >= len(coords):
                return True
            coords.pop()
            signs = [GeomLib.zCrossProduct(a, b, c) > 0 for a, b, c in zip(ListUtilities.rotate(coords, 2), ListUtilities.rotate(coords, 1), coords)]
            return all(signs) or not any(signs)
        return False

    @staticmethod
    def isHoled(obj):
        return (isinstance(obj, Polygon) and (0 < len(obj.interiors)))

    @staticmethod
    def isInFrontOf(viewpoint, bipointAsLinestring):
        if not isinstance(viewpoint, Point):
            raise IllegalArgumentTypeException(viewpoint, 'Point')
        if (not isinstance(bipointAsLinestring, LineString) or 
            (2 != len(bipointAsLinestring.coords))):
            raise IllegalArgumentTypeException(bipointAsLinestring, 'LineString of 2 points')
        a, b = bipointAsLinestring.coords
        ab = GeomLib.vector_to(a, b)
        n2ab = [ab[1], -ab[0]]
        ca = GeomLib.vector_to((viewpoint.x, viewpoint.y), a)
        return (GeomLib.dotProduct(n2ab, ca) < 0)

    @staticmethod
    def isLineal(obj):
        return isinstance(obj, (MultiLineString, LinearRing, LineString))

    @staticmethod
    def isMultipart(obj):
        return isinstance(obj, (GeometryCollection, MultiLineString, MultiPoint,
                                MultiPolygon))

    @staticmethod
    def isPolygonal(obj):
        return isinstance(obj, (MultiPolygon, Polygon))

    @staticmethod
    def isPuntal(obj):
        return isinstance(obj, (MultiPoint, Point))

    @staticmethod
    def normalizeRingOrientation(obj, ccw=True):
        # All exterior/interiors rings must be CCW oriented
        if (isinstance(obj, LineString) and obj.is_ring):
            _orientation = GeomLib.isCCW(obj)
            if (ccw == _orientation):
                return obj
            return LinearRing(reversed(obj.coords))

        elif isinstance(obj, Polygon):
            extRing = GeomLib.normalizeRingOrientation(obj.exterior, ccw)
            intRings = [GeomLib.normalizeRingOrientation(hole, not ccw) for hole in obj.interiors]
            return Polygon(extRing, intRings)

        elif isinstance(obj, MultiLineString):
            result = []
            for geom in obj.geoms:
                result.append(GeomLib.normalizeRingOrientation(geom, ccw))
            return MultiLineString(result)

        elif isinstance(obj, MultiPolygon):
            result = []
            for geom in obj.geoms:
                result.append(GeomLib.normalizeRingOrientation(geom, ccw))
            return MultiPolygon(result)

    @staticmethod
    def projectOntoStraightLine(p, line):
        # This method returns the projection 'projP' of point p on the 
        # given line but also the distance from p to the given line 
        # ('distFromPToLine') and the distance from the projection 
        # 'projP' to the first point of the input line. The input 
        # line is represented by two points.
        a, b = line
        ap = GeomLib.vector_to(a, p)
        unitAB = GeomLib.unitVector(a, b)
        orthoUnitAB = [-unitAB[1], unitAB[0]]
        distFromAToProjP = GeomLib.dotProduct(ap, unitAB)
        projP = [ a[0] + distFromAToProjP * unitAB[0], a[1] + distFromAToProjP * unitAB[1]]
        distFromPToLine = GeomLib.dotProduct(ap, orthoUnitAB)
        return [projP, distFromAToProjP, distFromPToLine]

    @staticmethod
    def removeHoles(obj):
        if isinstance(obj, Polygon):
            return Polygon(obj.exterior)
        elif isinstance(obj, MultiPolygon):
            return MultiPolygon([Polygon(p.exterior) for p in obj.geoms])
        return None
        # raise IllegalArgumentTypeException(obj, 'Polygon or MultiPolygon')

    @staticmethod
    def removeZCoordinate(obj):
        if GeomLib.isAShapelyGeometry(obj):
            return transform(lambda x, y, z=None: (x, y), obj)
        raise IllegalArgumentTypeException(obj, 'Shapely geometry')

    @staticmethod
    def reverseRingOrientation(obj):
        # if (isinstance(obj, LineString) and obj.is_ring):
        if isinstance(obj, LineString):
            return LineString(reversed(obj.coords))

        elif isinstance(obj, Polygon):
            extRing = GeomLib.reverseRingOrientation(obj.exterior)
            intRings = [GeomLib.reverseRingOrientation(hole) for hole in obj.interiors]
            return Polygon(extRing, intRings)

        elif isinstance(obj, MultiLineString):
            result = []
            for geom in obj.geoms:
                result.append(GeomLib.reverseRingOrientation(geom))
            return MultiLineString(result)

        elif isinstance(obj, MultiPolygon):
            result = []
            for geom in obj.geoms:
                result.append(GeomLib.reverseRingOrientation(geom))
            return MultiPolygon(result)

    @staticmethod
    def ringSignedArea(ring):
        # https://en.wikipedia.org/wiki/Shoelace_formula
        if (isinstance(ring, LineString)):  # and ring.is_ring):
            ring = ring.coords
            area = 0.0
            if len(ring) < 3:
                return area
            x1, y1 = ring[0][0:2]
            for xyz in ring[1:]:
                x2, y2 = xyz[0:2]
                area += (x1 * y2 - x2 * y1)
                x1, y1 = x2, y2
            return area / 2.0
        raise IllegalArgumentTypeException(ring, 'LineString')

    @staticmethod
    def splitSegmentAccordingToTheDistanceToViewpoint(segm, viewpt, dist):
        if not (isinstance(segm, LineString) and (2 == len(segm.coords))):
            raise IllegalArgumentTypeException(segm, '2-point LineString')
        if not isinstance(viewpt, Point):
            raise IllegalArgumentTypeException(viewpt, 'Point')
        p0, p1 = segm.coords

        # CHECK IF THE RESULTING POINT IS ON THE SEGMENT
        _maxDist = max(viewpt.distance(Point(p0)), viewpt.distance(Point(p1)))
        if ((dist < viewpt.distance(segm)) or (_maxDist < dist)):
            return []

        # LET'S SOLVE THE FOLLOWING SYSTEM:
        # x = x0 + (x1 - x0) * t
        # y = y0 + (y1 - y0) * t
        # (x - xv)**2 + (y - yv)**2 = dist**2
        # 0 <= t <= 1
        dx, dy = p1[0] - p0[0], p1[1] - p0[1]
        vx, vy = p0[0] - viewpt.x, p0[1] - viewpt.y

        a = (dx * dx + dy * dy)
        b = 2 * (dx * vx + dy * vy)
        c = (vx * vx + vy * vy - dist * dist)
        discr = b * b - 4 * a * c

        if (0 < discr):
            t1 = (-b + sqrt(discr)) / (2 * a)
            t2 = (-b - sqrt(discr)) / (2 * a)

            result = []
            if (0 <= t1 <= 1):
                result.append(Point([p0[0] + dx * t1, p0[1] + dy * t1]))
            if (0 <= t2 <= 1):
                result.append(Point([p0[0] + dx * t2, p0[1] + dy * t2]))
            return result

        elif (0 == discr):
            t0 = -b / (2 * a)
            if (0 <= t0 <= 1):
                return [ Point([p0[0] + dx * t0, p0[1] + dy * t0]) ]

        raise Exception('Unreachable instruction!')

    @staticmethod
    def toListOfLineStrings(obj):
        if GeomLib.isAShapelyGeometry(obj) and (not GeomLib.isPuntal(obj)):
            if isinstance(obj, LinearRing):
                return [LineString(obj.coords)]
            elif isinstance(obj, LineString):
                return [obj]
            elif isinstance(obj, Polygon):
                result = [LineString(obj.exterior.coords)]
                for r in obj.interiors:
                    result.append(LineString(r.coords))
                return result
            elif GeomLib.isMultipart(obj):
                result = []
                for g in obj.geoms:
                    result += GeomLib.toListOfLineStrings(g)
                return result
        return []

    @staticmethod
    def toListOfBipointsAsLineStrings(obj):
        result = []
        for geom in GeomLib.toListOfLineStrings(obj):
            prev = None
            for curr in geom.coords:
                if not prev is None:
                    result.append(LineString([prev, curr]))
                prev = curr
        return result
    
    @staticmethod
    def toListOfPolygons(obj):
        if GeomLib.isAShapelyGeometry(obj):
            if isinstance(obj, Polygon):
                return [obj]
            elif isinstance(obj, MultiPolygon):
                return obj.geoms
            elif isinstance(obj, GeometryCollection):
                result = []
                for g in obj.geoms:
                    result += GeomLib.toListOfPolygons(g)
                return result
        return []

    @staticmethod
    def toMultiPolygon(obj):
        if GeomLib.isAShapelyGeometry(obj):
            if isinstance(obj, Polygon):
                return MultiPolygon([obj])
            elif isinstance(obj, MultiPolygon):
                return obj
            elif isinstance(obj, GeometryCollection):
                return MultiPolygon(GeomLib.toListOfPolygons(obj))
        return MultiPolygon()

    @staticmethod
    def unitVector(a, b):
        ab = GeomLib.vector_to(a, b)
        invNorm = 1.0 / sqrt(ab[0] * ab[0] + ab[1] * ab[1])
        return [ab[0] * invNorm, ab[1] * invNorm]

    @staticmethod
    def vector_to(a, b):
        return [b[0] - a[0], b[1] - a[1]]

    @staticmethod
    def zCrossProduct(a, b, c):
        return (b[0] - a[0]) * (c[1] - b[1]) - (b[1] - a[1]) * (c[0] - b[0])



'''
Created on 1 avr. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from os import path
from geopandas import read_file


class TestUtils(object):
    '''
    classdocs
    '''

    @staticmethod
    def getDataSetFilename(dirname, basename):
        if path.exists('%s/%s' % (dirname, basename)):
            return '%s/%s' % (dirname, basename)
        elif path.exists('data/%s' % (basename)):
            return 'data/%s' % (basename)
        raise Exception('dataset not found: %s' % (basename)) 

    @staticmethod
    def loadDataSet(dirname, basename, driver=None):
        if driver is None:
            return read_file(TestUtils.getDataSetFilename(dirname, basename))
        return read_file(TestUtils.getDataSetFilename(dirname, basename), driver=driver)



'''
Created on 23 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
import itertools

from numpy import arccos, arctan2, array, cos, dot, flip, imag, log, matmul, ones, pi, polyadd, polyder, polymul, polyval, real, sin, sqrt, zeros
from numpy.linalg import cond, det, LinAlgError, solve, inv  # VARIANTE: from scipy.linalg import cond, det, inv, solve
from numpy.polynomial.polynomial import polyroots
from shapely.geometry import MultiPoint, Point

from t4gpd.commons.BoundingBox import BoundingBox
from t4gpd.commons.Epsilon import Epsilon
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.ellipse.EllipseLib import EllipseLib


class EllipticHullLib(object):
    '''
    classdocs
    '''

    def __init__(self, threshold=None, debug=False):
        '''
        Constructor
        '''
        self.threshold = threshold
        self.debug = debug

    def getMinimumAreaBoundingEllipse(self, qgsGeom):
        return self.getMinimumBoundingEllipse(qgsGeom, constraint='area')

    def getMinimumPerimeterBoundingEllipse(self, qgsGeom):
        return self.getMinimumBoundingEllipse(qgsGeom, constraint='perimeter')

    def getMinimumBoundingEllipse(self, geom, constraint='area'):
        if not GeomLib.isAShapelyGeometry(geom):
            raise IllegalArgumentTypeException(geom, 'Shapely geometry')
        coords = geom.convex_hull.exterior.coords[:-1]  # Clockwise order
        # coords = list(reversed(geom.convex_hull.exterior.coords[:-1]))  # Counter-clockwise order

        bbox = BoundingBox(geom)
        magnitude = 1.0 / max(bbox.width(), bbox.height())
        invMagnitude = 1.0 / magnitude

        if (self.debug):
            print('Nb sommets env. conv. : %d' % len(coords))
            print('\t[%s]' % '; '.join(['%.5f, %.5f' % (x, y) for x, y in coords])) 
            EllipticHullLib.__layerOfNodes(coords, 'convex hull', 'cross', 'black', '4.2')

        coords = EllipticHullLib.flatEdgesPruning(coords, self.threshold)
        if (self.debug):
            print('Nb sommets env. conv. simplifiee : %d' % len(coords))
            EllipticHullLib.__layerOfNodes(coords, 'simple convex hull', 'pentagon', 'red', '1.8')

        orderedIndices, centroid = EllipticHullLib.orderNodesAndIdentifyTheCentroid(coords)

        coords = EllipticHullLib.homotheticTransformation(EllipticHullLib.translateNodes(coords, [-centroid[0], -centroid[1]]), magnitude)

        if ('area' == constraint):
            nodes, centre, semiXAxis, semiYAxis, azim, methName = self.__getMinimumAreaBoundingEllipseAux(coords, orderedIndices)
        elif ('perimeter' == constraint):
            nodes, centre, semiXAxis, semiYAxis, azim, methName = self.__getMinimumPerimeterBoundingEllipseAux(coords, orderedIndices)
        else:
            raise Exception('The "constraint" parameter value must be chosen among [area, perimeter]!')
        centre = EllipticHullLib.translateNodes(EllipticHullLib.homotheticTransformation([centre], invMagnitude), centroid)[0]
        semiXAxis, semiYAxis = invMagnitude * semiXAxis, invMagnitude * semiYAxis 

        nodes = EllipticHullLib.translateNodes(EllipticHullLib.homotheticTransformation(nodes, invMagnitude), centroid)

        '''
        if (self.debug):
            semiAxis = semiXAxis if (semiXAxis >= semiYAxis) else semiYAxis
            p1 = Point((centre[0] - semiAxis * cos(azim), centre[1] - semiAxis * sin(azim)))
            p2 = Point((centre[0] + semiAxis * cos(azim), centre[1] + semiAxis * sin(azim)))

            memDriver = MemoryDriver('LineString', 'main axis', [], crs=QgsCoordinateReferenceSystem(2154))
            memDriver.addFeatures([ {'the_geom': [ p1, p2 ]} ])
            layer = memDriver.getVectorLayer()
            setLineSymbol(layer, color='green', penstyle='dot', width='0.35')
            structLayer = EllipticHullLib.__layerOfNodes(nodes, 'structuring nodes', 'circle', 'blue', '3.2')
            setLabelSymbol(structLayer, 'gid', overPoint=False, size='24', color='black')
            print('\ncentre = {:.5f}, {:.5f}, semiXAxis = {:.5f}, semiYAxis = {:.5f}, azim = {:.5f}, methName = {:s}'.format(centre[0], centre[1], semiXAxis, semiYAxis, azim, methName))
            print('\tnodes = [%s]' % (', '.join(['({:.5f}, {:.5f})'.format(x, y) for x, y in nodes])))
            print('\tarea = %.5f, perimeter = %.5f' % (pi * semiXAxis * semiYAxis, EllipseUtilities.getEllipsePerimeter(semiXAxis, semiYAxis)))
        '''

        return [Point(centre), semiXAxis, semiYAxis, azim, nodes, methName]

    def __getMinimumAreaBoundingEllipseAux(self, coords, orderedIndices, epsilon=Epsilon.EPSILON):
        ncoords = len(coords)
        allIndices = list(range(ncoords))

        if 3 <= ncoords:
            cnt = 0
            for indices in itertools.combinations(allIndices, 3):
                cnt += 1
                [A, B, C], pivots, centre, methName = self.__aux(coords, orderedIndices, indices)
                if EllipticHullLib.areInside(coords, centre, A, B, C, epsilon):
                    semiXAxis, semiYAxis, azim = EllipticHullLib.getSemiAxesAndAzim([A, B, C])   
                    return [pivots, centre, semiXAxis, semiYAxis, azim, methName]

        maxInvAreaMulSquarePi, solution = float('-inf'), None
        if 4 <= ncoords:
            cnt = 0
            for indices in list(itertools.combinations(allIndices, 4)) + list(itertools.combinations(allIndices, 5)):
                cnt += 1
                result = self.__aux(coords, orderedIndices, indices)
                if result is None:
                    continue
                [A, B, C], pivots, centre, methName = result
                currInvAreaMulSquarePi = 4 * A * C - B * B
                if (maxInvAreaMulSquarePi < currInvAreaMulSquarePi):
                    if EllipticHullLib.areInside(coords, centre, A, B, C, epsilon):
                        maxInvAreaMulSquarePi = currInvAreaMulSquarePi
                        solution = result
        if solution is not None:
            [A, B, C], pivots, centre, methName = solution
            semiXAxis, semiYAxis, azim = EllipticHullLib.getSemiAxesAndAzim([A, B, C])
            return [pivots, centre, semiXAxis, semiYAxis, azim, methName]
        raise Exception('__getMinimumAreaBoundingEllipseAux(...): Unreachable instruction!')

    def __getAllMinimumAreaBoundingEllipse(self, coords, orderedIndices, epsilon=Epsilon.EPSILON):
        ncoords = len(coords)
        allIndices = list(range(ncoords))
        solutions = []

        if 3 <= ncoords:
            cnt = 0
            for indices in list(itertools.combinations(allIndices, 3)) + list(itertools.combinations(allIndices, 4)) + list(itertools.combinations(allIndices, 5)):
                cnt += 1
                result = self.__aux(coords, orderedIndices, indices)
                if result is None:
                    continue
                [A, B, C], pivots, centre, methName = result
                semiXAxis, semiYAxis, azim = EllipticHullLib.getSemiAxesAndAzim([A, B, C])   
                if EllipticHullLib.areInside(coords, centre, A, B, C, epsilon):
                    solutions.append([pivots, centre, semiXAxis, semiYAxis, azim, methName])
        return solutions
        
    def __getMinimumPerimeterBoundingEllipseAux(self, coords, orderedIndices, epsilon=Epsilon.EPSILON):
        ncoords = len(coords)
        allIndices = list(range(ncoords))

        minPerim, solution = float('inf'), None
        if 3 <= ncoords:
            cnt = 0
            for indices in list(itertools.combinations(allIndices, 3)) + list(itertools.combinations(allIndices, 4)) + list(itertools.combinations(allIndices, 5)):
                cnt += 1
                result = self.__aux(coords, orderedIndices, indices)
                if result is None:
                    continue
                [A, B, C], pivots, centre, methName = result

                semiXAxis, semiYAxis, azim = EllipticHullLib.getSemiAxesAndAzim([A, B, C])   

                currPerim = EllipseLib.getEllipsePerimeter(semiXAxis, semiYAxis)
                if (minPerim > currPerim):
                    if EllipticHullLib.areInside(coords, centre, A, B, C, epsilon):
                        minPerim = currPerim
                        solution = [pivots, centre, semiXAxis, semiYAxis, azim, methName]

        if solution is not None:
            return solution
        raise Exception('__getMinimumPerimeterBoundingEllipseAux(...): Unreachable instruction!')

    def __aux(self, coords, orderedIndices, indices):
        # The polygon must be topologically valid:
        indices = sorted([orderedIndices[i] for i in indices])
        coords = [coords[i] for i in indices]
        nindices = len(indices)

        # The initial centering is by no means universal. It is 
        # necessary to refocus locally using an isobarycentre.
        if (3 <= nindices):
            centroid = (
                sum([coords[k][0] for k in range(nindices)]) / float(nindices),
                sum([coords[k][1] for k in range(nindices)]) / float(nindices)
                )
            coords = EllipticHullLib.translateNodes(coords, [-centroid[0], -centroid[1]])

        centre, quadraticCoeffs = None, None
        if (3 == nindices):
            quadraticCoeffs = EllipticHullLib.tri(coords[0], coords[1], coords[2])
            centre, methName = centroid, 'tri'
        elif (4 == nindices):
            centre, quadraticCoeffs, methName = EllipticHullLib.quadri(coords[0], coords[1], coords[2], coords[3])
        elif (5 == nindices):
            centre, quadraticCoeffs = EllipticHullLib.penta(coords[0], coords[1], coords[2], coords[3], coords[4])
            methName = 'penta'

        if centre is None:
            return None
        if quadraticCoeffs is None:
            quadraticCoeffs = EllipticHullLib.getQuadraticCoeffs(coords, (0, 0))
            if quadraticCoeffs is None:
                return None

        if (3 <= nindices):
            if (3 < nindices):
                centre = EllipticHullLib.translateNodes([centre], centroid)[0]
            coords = EllipticHullLib.translateNodes(coords, centroid)

        return [quadraticCoeffs, coords, centre, methName]

    @staticmethod
    def penta(p1, p2, p3, p4, p5):
        # The initial translation ensures that the origin of the coordinates system is an inner point of the ellipse
        # After such a translation it is legal to write the following linear system
        matrixA, vectorB = zeros((5, 5)), ones(5)
        for nline, (x, y) in enumerate([p1, p2, p3, p4, p5]):
            # A x2 + B xy + C y2 + D x + E y = 1
            matrixA[nline] = [x * x, x * y, y * y, x, y]

        # matrixA * vectorX = vectorB (with: vectorX = [a2, b2, c2, d2, e2])
        a2, b2, c2, d2, e2 = solve(matrixA, vectorB)

        # Based on formulas provided by: https://en.wikipedia.org/wiki/Ellipse
        discriminant = b2 * b2 - 4 * a2 * c2
        # ellipse (discriminant < 0); parabola (discriminant = 0); hyperbola (discriminant > 0) 
        if (discriminant < 0):
            x0 = (2.0 * c2 * d2 - b2 * e2) / discriminant
            y0 = (2.0 * a2 * e2 - b2 * d2) / discriminant
            magn = 1.0 / (1.0 - (a2 * x0 * x0 + b2 * x0 * y0 + c2 * y0 * y0 + d2 * x0 + e2 * y0))
            a, b, c = a2 * magn, b2 * magn , c2 * magn  
            return (x0, y0), [a, b, c]
        return None, None

    def getMinimumAreaBoundingEllipseValue(self, qgsGeom):
        _, semiXAxis, semiYAxis, _, _, _ = self.getMinimumAreaBoundingEllipse(qgsGeom)
        return pi * semiXAxis * semiYAxis

    def plotMinimumAreaBoundingEllipse(self, qgsGeom, npoints=40):
        centre, semiXAxis, semiYAxis, azim, _, _ = self.getMinimumAreaBoundingEllipse(qgsGeom)
        return [ EllipseLib.getEllipseContour(centre, semiXAxis, semiYAxis, azim, npoints) ]

    def plotMinimumAreaBoundingEllipseAxis(self, qgsGeom):
        centre, semiXAxis, semiYAxis, azim, _, _ = self.getMinimumAreaBoundingEllipse(qgsGeom)
        return EllipseLib.getEllipseMainAxis(centre, semiXAxis, semiYAxis, azim)

    def plotAllMinimumAreaBoundingEllipses(self, geom, npoints=40):
        if not GeomLib.isAShapelyGeometry(geom):
            raise IllegalArgumentTypeException(geom, 'Shapely geometry')
        coords = geom.convex_hull.exterior.coords[:-1]  # Clockwise order
        # coords = list(reversed(geom.convex_hull.exterior.coords[:-1]))  # Counter-clockwise order

        bbox = BoundingBox(geom)
        magnitude = 1.0 / max(bbox.width(), bbox.height())
        invMagnitude = 1.0 / magnitude

        coords = EllipticHullLib.flatEdgesPruning(coords, self.threshold)
        orderedIndices, centroid = EllipticHullLib.orderNodesAndIdentifyTheCentroid(coords)
        coords = EllipticHullLib.homotheticTransformation(EllipticHullLib.translateNodes(coords, [-centroid[0], -centroid[1]]), magnitude)
        solutions = self.__getAllMinimumAreaBoundingEllipse(coords, orderedIndices)

        for solution in solutions:
            _, centre, semiXAxis, semiYAxis, azim, methName = solution

            x0, y0 = EllipticHullLib.translateNodes(EllipticHullLib.homotheticTransformation([centre], invMagnitude), centroid)[0]
            semiXAxis, semiYAxis = invMagnitude * semiXAxis, invMagnitude * semiYAxis 
            # nodes = EllipticHullLib.translateNodes(EllipticHullLib.homotheticTransformation(nodes, invMagnitude), centroid)
            '''
            ellipticHull = STEllipse(x0, y0, semiXAxis, semiYAxis, azim, npoints, outputLayername='ellipse_%s' % methName).run()
            setSimpleOutlineFillSymbol(ellipticHull, color='red', width='0.2', penstyle='solid')
            '''

    @staticmethod
    def __flatEdgesPruning(p, q, r, s, threshold):
        pq = GeomLib.unitVector(p, q)
        qr = GeomLib.unitVector(q, r)
        rs = GeomLib.unitVector(r, s)
        a_pqr = arccos(GeomLib.dotProduct(pq, qr))
        a_qrs = arccos(GeomLib.dotProduct(qr, rs))
        if (threshold > a_pqr + a_qrs):
            return EllipticHullLib.intersect_line_line(p, q, r, s)
        return None

    @staticmethod
    def flatEdgesPruning(nodes, threshold=None):
        # IL FAUT QUE : nodes[0] == nodes[-1]
        result = [x for x in nodes]
        if (threshold is not None) and (0 < threshold):
            i, n = 0, len(result)
            while (i + 3 < n):
                # print('i = %d, n = %d, result[i:i + 4] : %s' % (i, n, result[i:i + 4]))
                p, q, r, s = result[i:i + 4]
                ans = EllipticHullLib.__flatEdgesPruning(p, q, r, s, threshold)
                if ans is None:
                    i += 1
                else:
                    del(result[i + 2])
                    result[i + 1] = ans
                    n = len(result)
        return result

    @staticmethod
    def applyLinearMapToNodes(linearMap, coords):
        return [dot(linearMap, coord) for coord in coords]

    @staticmethod
    def areInside(coords, centre, A, B, C, epsilon=Epsilon.EPSILON):
        # A x2 + B xy + C y2 <= 1
        one = 1.0 + epsilon
        coords = EllipticHullLib.translateNodes(coords, -array(centre))
        for x, y in coords:
            if (one < (A * x * x + B * x * y + C * y * y)):
                return False
        return True

    @staticmethod
    def getQuadraticCoeffs(coords, centre):
        # A x2 + B xy + C y2 = 1
        tcoords = EllipticHullLib.translateNodes(coords, -array(centre))

        minCondNumber, bestMatrix = float('inf'), None
        for a, b, c in itertools.combinations(tcoords, 3):
            matrix = [
                [a[0] * a[0], a[0] * a[1], a[1] * a[1]],
                [b[0] * b[0], b[0] * b[1], b[1] * b[1]],
                [c[0] * c[0], c[0] * c[1], c[1] * c[1]]
            ]
            currCondNumber = cond(matrix)
            if (minCondNumber > currCondNumber):
                minCondNumber = currCondNumber
                bestMatrix = matrix

        try:
            A, B, C = solve(bestMatrix, [1.0, 1.0, 1.0])
            return list(map(float, [A, B, C]))
        except LinAlgError as error:
            print('EllipticHullLib.getQuadraticCoeffs(...): %s' % error)
        return None

    @staticmethod
    def getSemiAxesAndAzim(quadraticCoeffs, epsilon=Epsilon.EPSILON):
        # A x2 + B xy + C y2 = 1
        A, B, C = quadraticCoeffs

        if Epsilon.equals(A, C, epsilon):
            if Epsilon.isZero(B):
                semiAxis = 1.0 / sqrt(A)
                return [semiAxis, semiAxis, float('nan')]
            azim = pi / 4.0
        else:
            azim = 0.5 * arctan2(B, A - C)

        cosAzim, sinAzim = cos(azim), sin(azim)
        sqCosAzim, sqSinAzim = cosAzim * cosAzim, sinAzim * sinAzim
        sinCos = cosAzim * sinAzim
        semiXAxis = 1.0 / sqrt(A * sqCosAzim + B * sinCos + C * sqSinAzim)
        semiYAxis = 1.0 / sqrt(C * sqCosAzim - B * sinCos + A * sqSinAzim)

        if (semiXAxis < semiYAxis):
            azim = (azim + pi / 2.0) if (azim < 0) else (azim - pi / 2.0)

        return [semiXAxis, semiYAxis, azim]

    @staticmethod
    def homotheticTransformation(coords, magnitude):
        if 1.0 == magnitude:
            return coords
        return [[magnitude * coord[0], magnitude * coord[1]] for coord in coords]

    @staticmethod
    def intersect_line_line(a, b, c, d):
        line1 = [a, GeomLib.vector_to(a, b)]
        line2 = [c, GeomLib.vector_to(c, d)]
        return GeomLib.intersect_line_line(line1, line2)

    @staticmethod
    def isParallelogram(a, b, c, d, epsilon=Epsilon.EPSILON):
        centre = (0.25 * (a[0] + b[0] + c[0] + d[0]), 0.25 * (a[1] + b[1] + c[1] + d[1]))
        a, b, c, d = EllipticHullLib.translateNodes([a, b, c, d], -array(centre))
        M21 = [[a[0], c[0]], [a[1], c[1]]]
        M22 = [[b[0], d[0]], [b[1], d[1]]]
        return (max(abs(det(M21)), abs(det(M22))) <= epsilon)

    @staticmethod
    def orderNodesAndIdentifyTheCentroid(coords):
        s, n, indices = coords, len(coords), list(range(len(coords)))

        edges = []
        for k in indices:
            dx = s[k][0] - s[k - 1][0]
            dy = s[k][1] - s[k - 1][1]
            magn = 1.0 / sqrt(dx * dx + dy * dy)
            edges.append((magn * dx, magn * dy))

        weights = [arccos(
            edges[k][0] * edges[(k + 1) % n][0] + 
            edges[k][1] * edges[(k + 1) % n][1]
            ) for k in indices]
        '''
        for w in weights:
            if w < 0:
                raise Exception('Each weight must be positive!')
        if not (Epsilon.equals(sum(weights), 2 * pi)):
            raise Exception('The sum of weights must be equal to 2pi!')
        ''' 
        invSumOfWeights = 1.0 / sum(weights)

        centroid = (
            sum([weights[k] * coords[k][0] for k in indices]) * invSumOfWeights,
            sum([weights[k] * coords[k][1] for k in indices]) * invSumOfWeights
            )

        orderedIndices = sorted(indices, key=lambda k: weights[k],
                                reverse=False)
        return [orderedIndices, centroid]

    @staticmethod
    def __parall(a, b, c, d):
        [m11, m12], [m21, m22] = inv([[a[0], b[0]], [a[1], b[1]]])
        centre = (0.0, 0.0)
        A = m11 * m11 + m21 * m21
        B = 2.0 * (m11 * m12 + m21 * m22)
        C = m12 * m12 + m22 * m22
        return [centre, [A, B, C]]
    
    @staticmethod
    def __other_quadri(a, b, c, d, epsilon):
        # m0 = [ (a[0] + b[0] + c[0] + d[0]) / 4.0, (a[1] + b[1] + c[1] + d[1]) / 4.0]
        P = [[a[0], b[0], c[0], d[0]], [a[1], b[1], c[1], d[1]]]

        M21 = [[a[0], c[0]], [a[1], c[1]]]
        M22 = [[b[0], d[0]], [b[1], d[1]]]

        # STEP 1
        if Epsilon.isZero(det(M21), epsilon):
            u1, v1 = float('inf'), float('inf')
        else:
            M21 = inv(M21)
            P1 = matmul(M21, P)
            if 0 < P1[0][1]:
                u1, v1 = P1[0][1], P1[1][1]
            else:
                u1, v1 = P1[0][3], P1[1][3]

        if Epsilon.isZero(det(M22), epsilon):
            u2, v2 = float('inf'), float('inf')
        else:
            M22 = inv(M22)
            P2 = matmul(M22, P)
            if 0 < P2[0][0]:
                u2, v2 = P2[0][0], P2[1][0]
            else:
                u2, v2 = P2[0][2], P2[1][2]
        
        # STEP 2
        if abs(u1 + v1 - 2) < abs(u2 + v2 - 2):
            u, v, M2, P = u1, v1, M21, P1
        else:
            u, v, M2, P = u2, v2, M22, P2

        # STEP 3
        if abs(log(u * (u + 1) * (1 + 3 * v - u))) > abs(log(v * (v + 1) * (1 + 3 * u - v))):
            u, v = v, u
            P = matmul([[0.0, 1.0], [1.0, 0.0]], P)
            M2 = matmul([[0.0, 1.0], [1.0, 0.0]], M2)

        du = u * (u + 1.0) * (1.0 + 3.0 * v - u)
        dv = v * (v + 1.0) * (1.0 + 3.0 * u - v)
        d2 = 2.0 / du
        a = [dv / du, ((v - 1.0) * ((1.0 + v) ** 2.0 - u * v) + (1.0 + 2.0 * v) * u ** 2.0) / du]
        b = [d2 * (v - u) * (2.0 * u * v + u + v - 1.0), d2 * ((1.0 - u) * ((1.0 + u) ** 2.0 - u * v) - (1.0 + 2.0 * u) * v ** 2.0)]
        c = [-1.0, 1.0]
        d = [-a[0], 1.0 - a[1]]

        den = polyadd(4.0 * polymul(a, c), -polymul(b, b))
        den2 = [0.0] + list(polymul(den, den))

        polt = [1.0, 0.0]
        m1 = polyadd(polymul(b, polt), -2.0 * polymul(c, d))
        m2 = polyadd(polymul(b, d), -2.0 * polymul(a, polt))

        num = polyadd(polymul(d, polymul(den, m1)), polymul(polt, polymul(den, m2)))
        num = polyadd(polymul(c, polymul(m2, m2)), num)
        num = polyadd(polymul(b, polymul(m1, m2)), num)
        num = polyadd(polymul(a, polymul(m1, m1)), num)
        num = polyadd(den2, -num)

        t6 = polyadd(2.0 * polymul(polyder(num), den), -5.0 * polymul(polyder(den), num))

        # DEBUG : FORCE THE MOST SIGNIFICANT FACTOR TO BE EQUAL TO 0 (EXACT EQUALITY)
        if 7 == len(t6):
            t6[0] = 0.0
        # t6 = roots(t6)
        t6 = polyroots(flip(t6, axis=0))

        v0, tt = float('inf'), None
        for t in t6:
            if Epsilon.isZero(abs(imag(t)), epsilon):
                t = real(t)
                if 1e-12 < polyval(den, t):
                    x, y = polyval(num, t), polyval(den, t)  
                    v = (x * x) / (y ** 5)
                    if v0 > v:
                        v0, tt = v, t

        if tt is None:
            # print('.')
            return [None, []]

        mm = array([[polyval(m1, tt)], [polyval(m2, tt)]]) / polyval(den, tt)

        for l in range(2):
            for k in range(4):
                P[l][k] = P[l][k] - mm[l][0]

        m = matmul(inv(M2), mm);

        dd = polyval(num, tt) / polyval(den2, tt)
        aa = polyval(a, tt) / dd
        bb = polyval(b, tt) / dd
        cc = polyval(c, tt) / dd;

        [m11, m12], [m21, m22] = M2

        aa1 = aa * m11 * m11 + bb * m11 * m21 + cc * m21 * m21
        bb1 = 2.0 * (aa * m11 * m12 + cc * m21 * m22) + bb * (m11 * m22 + m21 * m12)
        cc1 = aa * m12 * m12 + bb * m12 * m22 + cc * m22 * m22
        aa, bb, cc = float(aa1), float(bb1), float(cc1)

        centre = (real(m[0])[0], real(m[1])[0])
        return [centre, [aa, bb, cc]]

    @staticmethod
    def quadri(a, b, c, d, epsilon=Epsilon.EPSILON):
        if EllipticHullLib.isParallelogram(a, b, c, d, epsilon):
            return EllipticHullLib.__parall(a, b, c, d) + ['parall']
        return EllipticHullLib.__other_quadri(a, b, c, d, epsilon) + ['quadri']

    @staticmethod
    def rotateNodes(azim, coords):
        cosAzim, sinAzim = cos(azim), sin(azim)
        return [[coord[0] * cosAzim - coord[1] * sinAzim,
                    coord[0] * sinAzim + coord[1] * cosAzim] for coord in coords]

    @staticmethod
    def translateNodes(coords, t):
        if (0.0 == t[0]) and (0.0 == t[1]):
            return coords
        return [[coord[0] + t[0], coord[1] + t[1]] for coord in coords]

    @staticmethod
    def tri(a, b, c):
        '''
        Triangle a, b, c is transformed, via a linear map, into the equilateral triangle: 
        (1 0), (-1/2 V3/2), (-1/2 -V3/2)
        '''
        [m11, m12], [m21, m22] = matmul([[1.0, -0.5], [0.0, 0.5 * sqrt(3)]], inv([[a[0], b[0]], [a[1], b[1]]]))
        A = m11 * m11 + m21 * m21
        B = 2.0 * (m11 * m12 + m21 * m22)
        C = m12 * m12 + m22 * m22
        return A, B, C

    @staticmethod
    def __layerOfNodes(coords, label, symbolType, symbolColor, symbolSize):
        return GeoDataFrame({'geometry': MultiPoint(coords)}, crs='EPSG:2154')



'''
Created on 22 juin 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from warnings import warn

from geopandas.geodataframe import GeoDataFrame
from numpy import arcsin, array_equal, asarray, ceil, cos, dot, linspace, \
    meshgrid, multiply, pi, resize, sin, sqrt, vstack, where
from numpy.linalg import norm
from numpy.random import uniform
from pyvista import Sphere
from shapely.geometry.linestring import LineString
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.RotationLib import RotationLib
from t4gpd.commons.SphericalCartesianCoordinates import SphericalCartesianCoordinates
from t4gpd.pyvista.GeodeCiel import GeodeCiel
from t4gpd.pyvista.Icosahedron import Icosahedron


class RayCasting3DLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def fromThetaPhiToNCells(theta, phi): 
        if (3 <= theta) and (3 <= phi):
            # ~ return theta * ((phi - 1 - 2) * 2 + 2)
            return 2 * theta * (phi - 2)
        return 6

    @staticmethod
    def fromNCellsToThetaPhi(n_cells): 
        theta = ceil(sqrt(n_cells))
        phi = ceil(theta / 2) + 2
        return theta.astype(int), phi.astype(int), RayCasting3DLib.fromThetaPhiToNCells(theta, phi)

    @staticmethod
    def preparePanopticRays(nrays, method='geodeciel'):
        if ('pyvista' == method):
            theta, phi, _ = RayCasting3DLib.fromNCellsToThetaPhi(nrays)
            sphere = Sphere(radius=1.0, center=(0, 0, 0), direction=(0, 0, 1),
                            theta_resolution=theta, phi_resolution=phi)
            shootingDirs = asarray(sphere.cell_centers().points)
        elif ('geodeciel' == method):
            norecursions = GeodeCiel.fromNRaysToNRecursions(nrays)
            shootingDirs = asarray(GeodeCiel(norecursions).getRays())
        elif ('icosahedron' == method):
            norecursions = Icosahedron.fromNRaysToNRecursions(nrays)
            shootingDirs = asarray(Icosahedron(norecursions).getRays())
        elif ('basic' == method):
            angles = linspace(0, pi / 2, ceil(sqrt(nrays)).astype(int))
            lat, lon = meshgrid(angles, angles)
            lat, lon = lat.reshape(-1), lon.reshape(-1)
            cosLat = cos(lat) 
            x, y, z = cosLat * cos(lon), cosLat * sin(lon), sin(lat)
            xyz = vstack([x, y, z]).T
            shootingDirs = vstack([
                multiply(sign, xyz)
                for sign in [
                    (1, 1, 1), (-1, 1, 1), (1, -1, 1), (-1, -1, 1),
                    (1, 1, -1), (-1, 1, -1), (1, -1, -1), (-1, -1, -1)
                    ]
                ])
        elif ('MonteCarlo' == method):
            shootingDirs = RayCasting3DLib.preparePanopticRandomRays(nrays)
        else:
            raise IllegalArgumentTypeException(method, "'geodeciel', 'icosahedron', 'pyvista', 'MonteCarlo' or 'basic'")
        print(f"{method}:: Let's prepare {len(shootingDirs)} rays!")
        return shootingDirs

    @staticmethod
    def prepareOrientedRandomRays2(nrays, main_direction, openness=20):
        _, lon_n, lat_n = SphericalCartesianCoordinates.fromCartesianToSphericalCoordinates(*main_direction)
        openness = (pi * openness) / 180

        u = sqrt(uniform(low=0, high=1, size=nrays))
        v = uniform(low=0, high=2 * pi, size=nrays)

        # LONGITUDE, LATITUDE
        lon, lat = lon_n + openness * u * cos(v), lat_n + openness * u * sin(v)

        # FROM LONGITUDE AND LATITUDE TO CARTESIAN COORDINATES
        cosLat = cos(lat)
        x, y, z = cos(lon) * cosLat, sin(lon) * cosLat, sin(lat)

        return vstack([x, y, z]).T 

    @staticmethod
    def prepareOrientedRandomRays(nrays, main_direction, openness=20, method='MonteCarlo'):
        openness = (pi * openness) / 180
        cosOpenness = cos(openness)
        main_direction = main_direction / norm(main_direction)

        if ('MonteCarlo' == method):
            safetyFactor = 1.6
            nrays2 = int(safetyFactor * 2 * nrays / (1 - cos(openness)))
            rays = RayCasting3DLib.preparePanopticRandomRays(nrays2)
            dotProducts = dot(rays, main_direction)
            indices = where(dotProducts >= cosOpenness)[0]
            if (nrays > len(indices)):
                warn(f'In the Monte Carlo approach, safetyFactor should be greater than {safetyFactor}!')
            return resize(rays[indices,:], (nrays, 3))

        else:
            u = sqrt(uniform(low=0, high=1, size=nrays))
            v = uniform(low=0, high=2 * pi, size=nrays)

            # LONGITUDE, LATITUDE
            lon, lat = openness * u * cos(v), openness * u * sin(v)

            # FROM LONGITUDE AND LATITUDE TO CARTESIAN COORDINATES
            cosLat = cos(lat)
            x, y, z = cos(lon) * cosLat, sin(lon) * cosLat, sin(lat)

            inCoords = vstack([x, y, z]).T

            if array_equal([1, 0, 0], main_direction):
                outCoords = inCoords
            else:
                R = RotationLib.makeRotationMatrix([1, 0, 0], main_direction)
                outCoords = asarray([dot(R, inCoord) for inCoord in inCoords])
            return outCoords

    @staticmethod
    def preparePanopticRandomRays(nrays):
        # INSPIRED BY:
        # https://mathworld.wolfram.com/SpherePointPicking.html
        u, v = uniform(low=0, high=1, size=[2, nrays])

        # LONGITUDE, LATITUDE
        lon, lat = 2 * pi * u, arcsin(2 * v - 1)

        # FROM LONGITUDE AND LATITUDE TO CARTESIAN COORDINATES
        cosLat = cos(lat)
        x, y, z = cos(lon) * cosLat, sin(lon) * cosLat, sin(lat)

        return vstack([x, y, z]).T 

    @staticmethod
    def intoGeoDataFrame(rays, origin=[0, 0, 0]):
        rows = []
        for xyz in rays:
            rows.append({'geometry': LineString([
                origin,
                [origin[0] + xyz[0], origin[1] + xyz[1], origin[2] + xyz[2]]
                ])})
        return GeoDataFrame(rows)



'''
Created on 3 mars 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import arctan, tan
from shapely.geometry import LineString, Point
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class AbstractSunshineDuration(AbstractGeoprocess):
    '''
    classdocs
    '''

    def _getAllSunPositions(self, datetimes, maxElevation):
        result = []

        for dt in datetimes:
            alti, _ = self.sunModel.getSolarAnglesInRadians(dt)
            radDir = self.sunModel.getRadiationDirection(dt)
            if (0 < alti):
                rayLen = maxElevation / tan(alti)
                result.append((alti, radDir, rayLen))

        return result

    def _beingInTheSun(self, viewpoint, radDir, rayLen, sunAlti):
        # To avoid: "Inconsistent coordinate dimensionality"
        viewpoint = Point((viewpoint.x, viewpoint.y))
        remotePoint = Point((viewpoint.x + rayLen * radDir[0], viewpoint.y + rayLen * radDir[1]))
        sunRay = LineString([viewpoint, remotePoint])

        masksIds = list(self.masksSIdx.intersection(sunRay.bounds))
        for maskId in masksIds:
            mask = self.masksGdf.loc[maskId]
            maskGeom = mask.geometry
            maskElevation = mask[self.maskElevationFieldname]

            if sunRay.crosses(maskGeom):
                tmpGeom = maskGeom.intersection(sunRay)
                gc = tmpGeom.geoms if GeomLib.isMultipart(tmpGeom) else [tmpGeom]

                for geom in gc:
                    for vertex in geom.coords:
                        hitPoint = Point(vertex)
                        _dist = viewpoint.distance(hitPoint)
                        _angle = arctan(maskElevation / _dist)
                        if (_angle >= sunAlti):
                            return False

        return True



'''
Created on 18 dec. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import isnan, pi
from t4gpd.commons.ellipse.EllipseLib import EllipseLib
from t4gpd.commons.ellipse.EllipticHullLib import EllipticHullLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class EllipticityIndices(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, threshold=None, debug=False):
        '''
        Constructor
        '''
        self.mabe = EllipticHullLib(threshold, debug)

    def runWithArgs(self, row):
        geom = row.geometry
        geomArea = geom.area
        geomPerim = geom.length

        _, semiXAxis, semiYAxis, _, _, _ = self.mabe.getMinimumAreaBoundingEllipse(geom)

        mabeArea = pi * semiXAxis * semiYAxis
        if (isnan(mabeArea)) or (mabeArea < geomArea):
            raise Exception('Minimum-area bounding ellipse is %s (input geom area: %.1f)!' % (str(mabeArea), geomArea))
        mabePerim = EllipseLib.getEllipsePerimeter(semiXAxis, semiYAxis)

        flattening = min(semiXAxis, semiYAxis) / max(semiXAxis, semiYAxis) if (0.0 < max(semiXAxis, semiYAxis)) else None
        areaEllipDefect = geomArea / mabeArea if (0.0 < mabeArea) else None
        perimEllipDefect = mabePerim / geomPerim if (0.0 < geomPerim) else None

        return {
            'flattening': flattening,
            'a_elli_def': areaEllipDefect,
            'p_elli_def': perimEllipDefect,
            }        
